import argparse
import subprocess
import sys

from odm_sdk import GenestackException


def get_terminal_width():
    """
    Return terminal width in characters (defaults to 80).

    :return: terminal width
    :rtype: int
    """
    try:
        rows, columns = subprocess.check_output(['stty', 'size']).decode().split()
        return int(columns)
    except Exception as e:
        sys.stderr.write('Fail to get terminal size, use 80 as default: %s' % e)
        return 80


def isatty():
    """
    Return ``True`` if the file is connected to a tty device.

    :return: is a tty
    :rtype: bool
    """
    try:
        return sys.stdout.isatty()
    except AttributeError:
        return False


class GenestackArgumentParser(argparse.ArgumentParser):
    def parse_known_args(self, args=None, namespace=None):
        args, argv = super(GenestackArgumentParser, self).parse_known_args(args, namespace)
        if sum([bool(args.token), bool(args.user), bool(args.access_token)]) > 1:
            self.error('Exactly one of token, access_token or user should be specified')
        if args.pwd and not args.user:
            self.error('Password should not be specified without user')
        return args, argv


def make_connection_parser(user=None, password=None, host=None, token=None, access_token=None):
    """
    Creates an argument parser with the provided connection parameters.
    If one of ``email``, ``password`` or ``user`` is specified, they are used. Otherwise, the default
    identity from the local config file will be used.

    :param user: user alias or email
    :type user: str
    :param password: user password
    :type password: str
    :param host: host
    :type host: str
    :param token: API token string
    :type token: str
    :param access_token: access token string
    :type access_token: str
    :return: parser
    :rtype: argparse.ArgumentParser
    """
    parser = GenestackArgumentParser()
    group = parser.add_argument_group('connection')
    group.add_argument('-H', '--host', default=host, help='server host', metavar='<host>')
    group.add_argument('-u', '--user', dest='user', metavar='<user>', default=user,
                       help='user alias from settings or email')
    group.add_argument('-p', '--password', dest='pwd', default=password, metavar='<password>',
                       help='user password')
    group.add_argument('--token', dest='token', default=token, metavar='<api-token>',
                       help='API token to be used instead of the login and password')
    group.add_argument('--access-token', dest='access_token', default=access_token, metavar='<access-token>',
                       help='Access token or environment variable with its value to be used instead '
                            'of the login and password')
    group.add_argument('--debug', dest='debug', action='store_true',
                       help='include server stacktrace into error messages (implies --show-logs)')
    group.add_argument('--show-logs', dest='show_logs', action='store_true',
                       help='print application logs received from server to stdout')
    return parser


def get_user(args=None):
    """
    Returns the user corresponding to the provided arguments.
    If ``args`` is ``None``, uses :py:func:`~odm_sdk.make_connection_parser` to get arguments.

    :param args: result of commandline parse
    :type args: argparse.Namespace
    :return: user
    :rtype: settings.User
    """

    from .settings import config, User

    if args is None:
        args = make_connection_parser().parse_args()

    alias = args.user
    if args.token:
        return User(host=args.host, token=args.token)
    if args.access_token:
        return User(host=args.host, access_token=args.access_token)
    if not args.host and not args.pwd:
        if not alias and config.default_user:
            return config.default_user
        if alias in config.users:
            return config.users[alias]
    return User(email=alias, host=args.host, password=args.pwd, token=args.token)


def get_connection(args=None):
    """
    This is the same as :py:func:`~odm_sdk.get_user` . :py:meth:`~odm_sdk.settings.User.get_connection`
    Generally the fastest way to get an active connection.

    :param args: argument from :attr:`argparse.parse_args`
    :type args: argparse.Namespace
    :return: connection
    :rtype: odm_sdk.Connection
    """
    if args is None:
        args = make_connection_parser().parse_args()
    user = get_user(args)
    return user.get_connection(interactive=True, debug=args.debug, show_logs=args.show_logs)


def ask_confirmation(question, default=None):
    """
    Ask confirmation and return response as boolean value.
    Will be looping until user provides correct credentials.
    Does not check if program is attached to tty device.

    :param question: question to ask, without [y/n] suffix and question mark.
    :param default: default value for empty string. Can be ``'y'``, ``'n'``, and ``None``
    :return: whether user confirms
    :rtype: bool
    """
    if not isatty():
        raise GenestackException("Prompt cannot be called")

    assert default in ('y', 'n', None), 'Wrong default value, expect "n", "y" or None'
    question_suffix = '[%s/%s]' % tuple(x.upper() if x == default else x for x in 'yn')

    while True:
        text = input('%s %s? ' % (question, question_suffix)).strip().lower()
        if not text and default:
            text = default

        if text in ('y', 'yes'):
            return True
        if text in ('n', 'no'):
            return False
        print('Unexpected response please input "y[es]" or "n[o]"')


def interactive_select(items, message, to_string=None, selected=None):
    """
    Asks user to choose one of the items.
    Will be looping until user provides correct credentials.
    Does not check if program is attached to tty device.

    :param items: list of possible choices
    :param message: clarifying message
    :type to_string: function to convert choice to string
    :param selected: selected item by default
    :type selected: str
    :return:
    """
    to_string = to_string or str

    if selected and selected not in items:
        raise GenestackException('Selected item "%s" is not in list: %s' % (selected, items))

    about_default = ''

    while True:
        for i, option in enumerate(items, start=1):
            if option == selected:
                print('* ', end=' ')
                about_default = ' [%s]' % i
            else:
                print('  ', end=' ')
            print('%-2s %s' % ('%s)' % i, to_string(option)))

        raw_index = input('%s%s: ' % (message, about_default)).strip()

        if not raw_index and selected:
            return selected

        if not raw_index.isdigit():
            print('Wrong number: "%s"' % raw_index)
            continue

        item_index = int(raw_index) - 1

        if not 0 <= item_index < len(items):
            print('Number is not in list')
            continue
        return items[item_index]


def validate_constant(cls, key):
    constants = {v for k, v in cls.__dict__.items() if (not k.startswith("_") and isinstance(v, str))}
    return key in constants



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script creates new groups. It requires tsv file groups.tsv to be laying in the same directory
# as this script.
# groups.tsv should contain one group name per line, for example:
# group1
# group2

# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/1143242753/How+to+create+groups+and+add+existing+users+to+them+via+python+scripts
from __future__ import print_function, unicode_literals


import sys

from odm_sdk.utils import make_connection_parser, get_connection

from utils import colored, TsvReader, GREEN, RED


def get_groups(connection):
    return connection.application('genestack/groupsadmin-api').invoke('getData')['groups']


def create_group(name, connection):
    return connection.application('genestack/groupsadmin-api').invoke('createGroup', name)


def check_group_exists(groupname, groups):
    for group in groups:
        if group['name'] == groupname:
            return group['accession']
    return None


def create_groups(groups_to_create, existing_groups, connection):
    for groupname in groups_to_create:
        accession = check_group_exists(groupname, existing_groups)
        if accession:
            print(colored("Group %s already exists with accession %s"
                          % (groupname, accession),
                          RED))
        else:
            group_properties = create_group(groupname, connection)
            print(colored("Group %s was created with accession %s"
                          % (groupname, group_properties['accession']), GREEN))


def read_groups(args):
    groups_to_create = []
    with TsvReader(args.file_with_groups) as groupsfile:
        for row in groupsfile:
            try:
                group = row[0].strip()
                if len(group) == 0:
                    raise IndexError
                groups_to_create.append(group)
            except IndexError:
                continue
        if len(groups_to_create) == 0:
            print(colored('Error: At least one group should present in the file groups.tsv',
                          RED))
            sys.exit()
        return groups_to_create


def main():
    parser = make_connection_parser()
    parser.add_argument('--file_with_groups', metavar='<file_with_groups>', default="groups.tsv",
                        help='create groups')
    connection = get_connection()
    args = parser.parse_args()

    existing_groups = get_groups(connection)
    groups_to_create = read_groups(args)
    create_groups(groups_to_create, existing_groups, connection)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script adds token to the current user.

from __future__ import print_function, unicode_literals
from odm_sdk import make_connection_parser, get_connection


def create_token(connection, token_name, token_content):
    return connection.application('genestack/profile-api').invoke('createApiTokenDirectly', token_name, token_content)


def get_args():
    parser = make_connection_parser()
    parser.add_argument('--token-name', default="Token", help='Token name for creation')
    parser.add_argument('--token-content', required=True, help='Exact token to be created')
    return parser.parse_args()


def main():
    args = get_args()
    connection = get_connection(args)
    create_token(connection, args.token_name, args.token_content)


if __name__ == "__main__":
    main()



import os
import platform
from copy import deepcopy
from xml.dom.minidom import getDOMImplementation, parse

import sys

from odm_sdk import GenestackException
from odm_sdk.settings.user import User
from odm_sdk.utils import ask_confirmation

_PASSWORD_KEYRING = 'Genestack SDK'
_TOKEN_KEYRING = 'Genestack SDK (token)'

_SETTING_FILE_NAME = 'genestack.xml'
_SETTINGS_DIR = '.genestack'


class Config(object):
    def __init__(self):
        self.__default_user = None
        self.__users = {}
        self.store_raw = None
        self.store_raw_session = None

    def get_settings_folder(self):
        if platform.system() == 'Windows':
            import ctypes.wintypes

            # http://stackoverflow.com/a/3859336/1310066 26 is Roaming folder
            buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
            ctypes.windll.shell32.SHGetFolderPathW(0, 26, 0, 0, buf)
            path = os.path.join(str(buf.value), _SETTINGS_DIR)
        else:
            path = os.path.join(os.path.expanduser('~/'), _SETTINGS_DIR)
        return path

    def get_settings_file(self):
        return os.path.join(self.get_settings_folder(), _SETTING_FILE_NAME)

    @property
    def default_user(self):
        return deepcopy(self.__default_user)

    @property
    def users(self):
        return deepcopy(self.__users)

    def remove_user(self, user):
        del self.__users[user.alias]
        self.save()
        try:
            import keyring
        except ImportError:
            return
        try:
            if keyring.get_password(_PASSWORD_KEYRING, user.alias):
                keyring.delete_password(_PASSWORD_KEYRING, user.alias)
            if keyring.get_password(_TOKEN_KEYRING, user.alias):
                keyring.delete_password(_TOKEN_KEYRING, user.alias)
        except Exception as e:
            sys.stderr.write('Error while deleting user password for %s: %s\n' % (user.alias, e))

    def add_user(self, user, save=True):
        if not user.alias:
            raise GenestackException('Cannot add user without alias')
        if user.alias in self.__users:
            raise GenestackException('User alias "%s" already exists' % user.alias)
        self.__users[user.alias] = user
        if len(self.__users) == 1:
            self.set_default_user(user, save=False)
        if save:
            self.save()

    def set_default_user(self, user, save=True):
        if user.alias not in self.__users:
            raise GenestackException('User "%s" does not exist' % user.alias)
        if not self.default_user or user.alias != self.default_user.alias:
            self.__default_user = user
        if save:
            self.save()

    def load(self):
        config_path = self.get_settings_file()

        if not os.path.exists(config_path):
            return

        def get_text(parent, tag):
            elements = parent.getElementsByTagName(tag)
            if len(elements) == 1:
                return elements[0].childNodes[0].data.strip()
            return None

        with open(config_path) as f:
            dom = parse(f)

            users = dom.getElementsByTagName('user')
            for user in users:
                alias = get_text(user, 'alias')
                host = get_text(user, 'host')
                email = get_text(user, 'email')
                password = get_text(user, 'password')
                token = get_text(user, 'token')
                if not password:
                    try:
                        import keyring
                        password = keyring.get_password(_PASSWORD_KEYRING, alias)
                    except Exception as e:
                        sys.stderr.write('Fail to load password for alias "%s": %s\n' % (alias, e))
                if not token:
                    try:
                        import keyring
                        token = keyring.get_password(_TOKEN_KEYRING, alias)
                    except Exception as e:
                        sys.stderr.write('Fail to load token for alias "%s": %s\n' % (alias, e))
                self.add_user(User(email, alias=alias, host=host, password=password,
                                   token=token), save=False)

        default_user_alias = get_text(dom, 'default_user')
        store_raw_text = get_text(dom, 'store_raw')
        if store_raw_text == 'True':
            self.store_raw = True
        elif store_raw_text == 'False':
            self.store_raw = False
        try:
            default_user = self.__users[default_user_alias]
            self.set_default_user(default_user, save=False)
        except KeyError:
            raise GenestackException('Cannot set find user: "%s" is not present in config users' % default_user_alias)

    def change_password(self, alias, password):
        user = self.__users[alias]
        user.password = password
        self.save()

    def change_token(self, alias, token):
        user = self.__users[alias]
        user.token = token
        self.save()

    def save(self):
        settings_folder = self.get_settings_folder()
        if not os.path.exists(settings_folder):
            os.makedirs(settings_folder)
        config_path = self.get_settings_file()

        dom = getDOMImplementation()
        document = dom.createDocument(None, 'genestack', None)
        top = document.documentElement
        users_element = document.createElement('users')
        top.appendChild(users_element)

        for user in self.__users.values():
            if not user.alias:
                continue

            user_element = document.createElement('user')
            users_element.appendChild(user_element)
            alias_element = document.createElement('alias')
            alias_element.appendChild(document.createTextNode(user.alias))
            user_element.appendChild(alias_element)

            if user.email:
                email_element = document.createElement('email')
                email_element.appendChild(document.createTextNode(user.email))
                user_element.appendChild(email_element)

            if user.host:
                host_element = document.createElement('host')
                host_element.appendChild(document.createTextNode(user.host))
                user_element.appendChild(host_element)

            if user.password:
                try:
                    self._store_value_securely(_PASSWORD_KEYRING, user.alias, user.password)
                except Exception:
                    self._store_value_insecurely(user.password, document, user_element, 'password')
            if user.token:
                try:
                    self._store_value_securely(_TOKEN_KEYRING, user.alias, user.token, )
                except Exception:
                    self._store_value_insecurely(user.token, document, user_element, 'token')

        if self.default_user:
            default_user_element = document.createElement('default_user')
            top.appendChild(default_user_element)
            default_user_element.appendChild(document.createTextNode(self.default_user.alias))
        if self.store_raw is not None:
            store_raw_element = document.createElement('store_raw')
            top.appendChild(store_raw_element)
            store_raw_element.appendChild(document.createTextNode(str(self.store_raw)))
        with open(config_path, 'w') as f:
            document.writexml(f, indent='', addindent='    ', newl='\n')

    def _store_value_insecurely(self, value, document, parent, element_name):
        """
        Store value in XML config file.

        :param value: value to be stored
        :type value: basestring
        :param document: document
        :type document: xml.dom.minidom.Document
        :param parent: parent element
        :type parent: xml.dom.minidom.Element
        :param element_name: name of the element
        :type element_name: basestring
        :return: None
        """
        if self.store_raw is not None:
            save_to_file = self.store_raw
        elif self.store_raw_session is not None:
            save_to_file = self.store_raw_session
        else:
            try:
                save_to_file = ask_confirmation(
                    'Do you want to store sensitive data in config file in plain text?',
                    default='n')
            except KeyboardInterrupt:
                save_to_file = False
            try:
                self.store_raw = ask_confirmation('Set this as default behaviour?', default='y')
            except KeyboardInterrupt:
                self.store_raw_session = save_to_file
        if save_to_file:
            value_element = document.createElement(element_name)
            value_element.appendChild(document.createTextNode(value))
            parent.appendChild(value_element)
        else:
            sys.stderr.write('"%s" has not been saved to config file\n' % element_name)

    def _store_value_securely(self, service_name, username, secret_value):
        """
        Save value in security vault.

        :param service_name: key for storage
        :type service_name: basestring
        :param username: user alias
        :type username: basestring
        :param secret_value: value to be stored
        :type secret_value: basestring
        :return: None
        :raises Exception if not able to store value
        """
        try:
            import keyring
            keyring.set_password(service_name, username, secret_value)
        except Exception as e:
            if not self.store_raw:
                sys.stderr.write('Cannot store in secure storage: %s\n' % e)
            raise e

config = Config()
config.load()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script creates new users. It requires tsv file users.tsv to be laying in the same directory
# as this script.
# users.tsv should contain user's email and name and optionally password separated by tab per line,
# for example:
# alice@alphacorp.com	Alice	pwdAlice123
# bob@alphacorp.com	Bob

# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/1074823180/How+to+create+users+via+python+script
from __future__ import print_function, unicode_literals

import random
import re
import string
import sys
import time

from odm_sdk import GenestackServerException
from odm_sdk.utils import make_connection_parser, get_connection

from odm_sdk.scripts.utils import colored, TsvReader, GREEN, RED, BLUE


def create_user(connection, user):
    email = user['email']
    pwd = user['pwd']
    name = user['name']
    try:
        connection.application('genestack/usersadmin-api').invoke('createUser', email, pwd, name)
        print(colored("%s\t%s\t%s" % (email, pwd, name), GREEN))
    except GenestackServerException as e:
        p = re.compile('is already taken')
        result = p.search(e.message)
        if result is not None:
            print(colored("E-mail %s is already taken" % email, BLUE))
        else:
            raise e


def mkpasswd(length=8, digits=2, upper=2, lower=2):
    random.seed(time.time())
    password = (
            [random.choice(string.digits) for _ in range(digits)] +
            [random.choice(string.ascii_uppercase) for _ in range(upper)] +
            [random.choice(string.ascii_lowercase) for _ in range(lower)] +
            [random.choice(string.ascii_letters) for _ in range(length - digits - upper - lower)]
    )
    return "".join(random.sample(password, len(password)))


def read_users_from_file(args):
    users = []
    incorrect_rows = []
    with TsvReader(args.file_with_users) as usersfile:
        for row in usersfile:
            try:
                email = row[0].strip()
                name = row[1].strip()
                pwd = row[2].strip()
                if len(email) == 0 or len(name) == 0:
                    raise IndexError
                if len(pwd) == 0:
                    pwd = mkpasswd()
                users += [
                    {'email': email, 'pwd': pwd, 'name': name}
                ]
            except IndexError:
                incorrect_rows.append(row)
    if len(incorrect_rows) > 0:
        print(colored("Error: each line should contain email and user's name separated by "
                      "tab. \nReceived: %s. \nCorrect data and rerun script" % incorrect_rows,
                      RED))
        sys.exit()
    if len(users) == 0:
        print(colored('Error: At least one user should present in the file users.tsv',
                      RED))
        sys.exit()
    return users


def main():
    parser = make_connection_parser()
    parser.add_argument('--file_with_users', metavar='<file_with_users>', default="users.tsv",
                        help='create users')
    connection = get_connection()
    args = parser.parse_args()

    users = read_users_from_file(args)

    for user in users:
        create_user(connection, user)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function

import sys

from odm_sdk import (Application, FilesUtil, GenestackException,
                              GenestackServerException, SpecialFolders,
                              GroupsUtil, ShareUtil)

from . import json_validation
from .dictionary_search import DictionarySearch
from .format import colored, print_result, Color


# permission name to look for in "not enough permissions" server error
TEMPLATE_PERM_NAME = 'MANAGE_TEMPLATES'


class TemplateImporter(Application):
    """
    Class for uploading templates and dictionaries.
    """

    IMPORT_TEMPLATE_FOLDER_NAME = 'Import Templates'
    APPLICATION_ID = 'genestack/study-metainfotemplateeditor'

    def __init__(self, connection):
        super(TemplateImporter, self).__init__(connection)

        self._files_util = FilesUtil(connection)
        self._share_util = ShareUtil(connection)
        self._dictionary_search = DictionarySearch(connection)
        self._groups_util = GroupsUtil(connection)
        self._import_template_folder = self._get_import_template_folder()

    def invoke(self, method, *params):
        try:
            return super(TemplateImporter, self).invoke(method, *params)
        except GenestackServerException as e:
            if TEMPLATE_PERM_NAME in e.message:
                print("TIP: to create or update metainfo template, user has to "
                      "have 'Set up templates' permission; request it from your "
                      "administrator", file=sys.stderr)
            raise e

    def create_from_config(self, config_path):
        config = json_validation.load_json(config_path, schema_name='config_schema.json')
        self._create_from_config(**config)

    def _create_from_config(
            self, template_path, template_name,
            replace=False, mark_default=False
    ):

        template_accession = self._create_template(template_name, template_path, replace=replace)
        if template_accession is None:
            return

        if mark_default:
            self._mark_as_default(template_accession)
            # it would be nice to display the organization's name but as of 21.08.2020 API
            # method for doing this has been removed
            print('Template "%s" is marked as the Default Template'
                  % template_accession)

    def _create_template(self, name, json_path, replace):
        content = json_validation.load_json(json_path, schema_name='template_schema.json')
        accession = self._find_existing_template(name)

        if accession is None:
            print('No template with name "%s" found, creating a new template' %
                  colored(name, Color.BLUE))
            accession = self.invoke('createTemplateFile', name)

            print_result(accession, name, 'Template')
        elif replace:
            print('Template %s / %s already exists and will be reused' %
                  (colored(accession, Color.GREEN), colored(name, Color.BLUE)))
        else:
            print(
                "Template '%s'(%s) already exists in '%s', use '\"replace\": true' in the config "
                "to update existing or choose other name" %
                (name, accession, self._import_template_folder)
            )
            return None

        self._put_content(accession, content)
        return accession

    def _check_is_template(self, name, accession):
        if not accession:
            return False
        try:
            self.invoke('loadTemplate', accession)
        except GenestackException:
            print(
                'File "%s" (%s) has been found but is not a template' % (name, accession)
            )
            return False
        return True

    def _find_existing_template(self, name):
        def find(parent):
            return self._files_util.find_application_page_file_by_name(name, parent)

        folders_to_search = [self._files_util.get_public_folder(),
                             self._import_template_folder,
                             self._files_util.get_special_folder(SpecialFolders.CREATED)
                             ]
        for folder in folders_to_search:
            accession = find(folder)
            if self._check_is_template(name, accession):
                return accession
        return None

    def _get_import_template_folder(self):
        created_folder = self._files_util.get_special_folder(SpecialFolders.CREATED)
        return self._files_util.get_folder(
            created_folder, 'Data samples', self.IMPORT_TEMPLATE_FOLDER_NAME, create=True
        )

    def _put_content(self, accession, fields):
        def process_field(field):
            new_field = field.copy()
            dictionary_name = new_field.pop('dictionaryName', None)
            dictionary_path = new_field.pop('dictionaryPath', None)
            new_field['dictionaryAccession'] = self._find_dict(dictionary_path, dictionary_name)
            return new_field

        processed_fields = map(process_field, fields)

        types = {}
        for item in processed_fields:
            data_type = item['dataType']
            types.setdefault(data_type, []).append(item)

        for data_type, data_type_items in types.items():
            self.invoke('setKeyInfosForFileKind', accession, data_type, data_type_items)

    def _find_dict(self, path, name):
        if not name:
            return None

        dict_accession = self._dictionary_search.find_dictionary(path, name)
        if not dict_accession:
            raise GenestackException(
                "Can't find dictionary '%s' in '%s'" % (name, path)
            )
        return dict_accession

    def _mark_as_default(self, accession):
        self.invoke('setDefaultTemplate', accession)



import json
import os
import sys
import urllib
from io import FileIO
from urllib.parse import urlsplit
import requests
from requests import HTTPError, RequestException

from odm_sdk import (GenestackAuthenticationException, GenestackConnectionFailure,
                              GenestackException, GenestackResponseError, GenestackServerException,
                              GenestackVersionException, __version__)
from odm_sdk.chunked_upload import upload_by_chunks
from odm_sdk.utils import isatty


class Response(object):
    """Represents response from Genestack server."""

    def __init__(self, data):
        self._data = data

    @property
    def error(self):
        return self._data.get('error')

    @property
    def error_stack_trace(self):
        return self._data.get('errorStackTrace')

    @property
    def log(self):
        return self._data['log']

    @property
    def result(self):
        return self._data['result']

    @property
    def trace(self):
        return self._data.get('trace')

    @property
    def elapsed_microseconds(self):
        return self._data.get('elapsedMicroseconds')


class Connection(object):
    """
    A class to handle a connection to a specified Genestack server.
    Instantiating the class does mean you are logged in to the server.
    To do so, you need to call the :py:meth:`~odm_sdk.Connection.login` method.
    """

    def __init__(self, server_url, debug=False, show_logs=False):
        """
        :param server_url: server url
        :type server_url: str
        :param debug:  will print additional traceback from application
        :type debug: bool
        :param show_logs: will print application logs (received from server)
        :type show_logs: bool
        """
        self.server_url = server_url
        self.session = requests.session()
        self.debug = debug
        self.show_logs = show_logs

    def __del__(self):
        try:
            self.logout()
        except Exception:
            # fail silently
            pass

    def whoami(self):
        """
        Return user email.

        :return: email
        :rtype: str
        """
        return self.application('genestack/signin').invoke('whoami')

    @staticmethod
    def _handle_inactive_user_auth(auth_result):
        """
        Check authentication response and throw descriptive exception for inactive users.

        :param auth_result: authentication response
        """
        if auth_result.get('userDeactivated'):
            raise GenestackAuthenticationException(
                "User's Genestack account was deactivated."
                " Contact your organization administrator")
        elif auth_result.get('userNotConfirmed'):
            raise GenestackAuthenticationException(
                "User's Genestack account was not confirmed")

    def login(self, email, password):
        """
        Attempt a login on the connection with the specified credentials.
        Raises an exception if the login fails.

        :param email: email
        :type email: str
        :param password: password
        :type password: str
        :rtype: None
        :raises: :py:class:`~odm_sdk.GenestackAuthenticationException` if login failed
        """
        logged = self.application('genestack/signin').invoke('authenticate', email, password)
        Connection._handle_inactive_user_auth(logged)
        if not logged['authenticated']:
            hostname = urlsplit(self.server_url).hostname
            raise GenestackAuthenticationException(
                'Fail to login with "%s" to "%s"' % (email, hostname))

    def login_by_token(self, token):
        """
        Attempt a login on the connection with the specified token.
        Raises an exception if the login fails.

        :param token: token
        :rtype: None
        :raises: :py:class:`~odm_sdk.GenestackAuthenticationException` if login failed
        """
        logged = self.application('genestack/signin').invoke('authenticateByApiToken', token)
        Connection._handle_inactive_user_auth(logged)
        if not logged['authenticated']:
            hostname = urlsplit(self.server_url).hostname
            raise GenestackAuthenticationException('Fail to login by token to %s' % hostname)

    def login_by_access_token(self, access_token):
        """
        Attempt a login on the connection with the specified access token.
        Raises an exception if the login fails.

        :param access_token: OAuth access token
        :type access_token: str
        :rtype: None
        :raises: :py:class:`~odm_sdk.GenestackAuthenticationException` if login failed
        """
        access_token = os.getenv(access_token, access_token)
        self.session.headers.update({'Authorization': f'Bearer {access_token}'})
        logged = self.application('genestack/signin').invoke('authenticateOAuthAccessToken', access_token)
        Connection._handle_inactive_user_auth(logged)
        if not logged['authenticated']:
            self.session.headers.pop('Authorization')
            hostname = urlsplit(self.server_url).hostname
            raise GenestackAuthenticationException('Fail to login by token to %s' % hostname)

    def logout(self):
        """
        Logout from server.

        :rtype: None
        """
        self.application('genestack/signin').invoke('signOut')

    def perform_request(self, path, data='', follow=True, headers=None):
        """
        Perform an HTTP request to Genestack server.

        Connects to remote server and sends ``data`` to an endpoint ``path``
        with additional ``headers``.

        :param str path: URL path (endpoint) to be used (concatenated with
                     ``self.server_url``).
        :param dict|file|str data: dictionary, bytes, or file-like object to send in the body
        :param bool follow: should we follow a redirection (if any)
        :param dict[str, str] headers: dictionary of additional headers; list of pairs is
                        supported too until v1.0 (for backward compatibility)
        :return: response from server
        :rtype: Response
        """
        _headers = {'gs-extendSession': 'true'}

        if headers:
            _headers.update(headers)
        try:
            response = self.session.post(
                self.server_url + path, data=data, headers=_headers,
                allow_redirects=follow,
                timeout=int(os.environ.get("GENESTACK_CLIENT_TIMEOUT"))
                if os.environ.get("GENESTACK_CLIENT_TIMEOUT", None) else None)
            if response.status_code == 401:
                raise GenestackAuthenticationException('Authentication failure')
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise GenestackResponseError(*e.args)
            try:
                return Response(response.json())
            except ValueError as e:
                raise GenestackException('Cannot parse content: %s' % e)

        except RequestException as e:
            raise GenestackConnectionFailure(str(e))

    def application(self, application_id):
        """
        Returns an application handler for the application with the specified ID.

        :param application_id: Application ID.
        :type application_id: str
        :return: application class
        :rtype: Application
        """
        return Application(self, application_id)

    def __repr__(self):
        return 'Connection("%s")' % self.server_url

    def get_request(self, path, params=None, follow=True):
        # This request also serves for download method of an application
        # It is possible that next request that uses same connection will fail,
        # that why we create new connection every time instead of reuse self.session
        return requests.get(
            url=self.server_url + path,
            params=params,
            allow_redirects=follow,
            cookies=self.session.cookies
        )

    def post_multipart(self, path, data=None, files=None, follow=True):
        return self.session.post(
            url=self.server_url + path,
            data=data,
            files=files,
            allow_redirects=follow,
        )


class Application(object):
    """
    Create a new application instance for the given connection.
    The connection must be logged in to call the application's methods.
    The application ID can be specified either as an argument to the class constructor
    or by overriding the ``APPLICATION_ID`` attribute in a child class.
    """

    APPLICATION_ID = None

    def __init__(self, connection, application_id=None):
        if application_id and self.APPLICATION_ID:
            raise GenestackException(
                "Application ID specified both as argument and as class variable"
            )
        self.application_id = application_id or self.APPLICATION_ID
        if not self.application_id:
            raise GenestackException('Application ID was not specified')

        self.connection = connection

        # validation of application ID
        if len(self.application_id.split('/')) != 2:
            raise GenestackException(
                'Invalid application ID, expect "{vendor}/{application}" got: %s' % self.application_id
            )

    def __invoke(self, path, post_data, trace=None):
        headers = {}
        if trace:
            headers['Genestack-Trace'] = 'true'
        response = self.connection.perform_request(path, post_data, headers=headers)

        if response.error is not None:
            raise GenestackServerException(
                response.error, path, post_data,
                debug=self.connection.debug,
                stack_trace=response.error_stack_trace
            )

        if response.log and (self.connection.show_logs or self.connection.debug):
            message = '\nLogs:\n' + '\n'.join(
                item['message'] + ('\n' + item['stackTrace'] if item.get('stackTrace') else '') for item in response.log
            )
            print(message)

        return response

    def get_response(self, method, params=None, trace=True):
        """
        Invoke one of the application's public Java methods and return Response object.
        Allow to access to logs and traces in code,
        if you need only result use :py:meth:`~odm_sdk.Application.invoke`

        :param method: name of the public Java method
        :type method: str
        :param params: arguments that will be passed to the Java method.
                       Arguments must be JSON-serializable.
        :type params: tuple
        :param trace: request trace from server
        :type trace: bool
        :return: Response object
        :rtype: Response
        """
        if not params:
            params = []

        post_data = json.dumps(params)
        path = '/application/invoke/%s/%s' % (self.application_id, urllib.parse.quote(method))

        # there might be present also self.__invoke(path, post_data)['log'] -- show it?
        return self.__invoke(path, post_data, trace=trace)

    def invoke(self, method, *params):
        """
        Invoke one of the application's public Java methods.

        :param method: name of the public Java method
        :type method: str
        :param params: arguments that will be passed to the Java method.
                       Arguments must be JSON-serializable.
        :return: JSON-deserialized response
        """
        return self.get_response(method, params).result

    def upload_chunked_file(self, file_path):
        return upload_by_chunks(self, file_path)

    def upload_file(self, file_path, token):
        """
        Upload a file to the current Genestack instance.
        This action requires a special token that can be generated by the application.

        :param file_path: path to existing local file.
        :type file_path: str
        :param token: upload token
        :type file_path: str
        :rtype: None
        """
        if isatty():
            progress = TTYProgress()
        else:
            progress = DottedProgress(40)

        file_to_upload = FileWithCallback(file_path, 'rb', progress)
        filename = os.path.basename(file_path)
        path = '/application/upload/%s/%s/%s' % (
            self.application_id, token, urllib.parse.quote(filename)
        )
        return self.__invoke(path, file_to_upload).result


class FileWithCallback(FileIO):
    def __init__(self, path, mode, callback):
        FileIO.__init__(self, path, mode)
        self.seek(0, os.SEEK_END)
        self.__total = self.tell()
        self.seek(0)
        self.__callback = callback

    def __len__(self):
        return self.__total

    def read(self, size=None):
        data = FileIO.read(self, size)
        self.__callback(os.path.basename(self.name), len(data), self.__total)
        return data


class TTYProgress(object):
    def __init__(self):
        self._seen = 0.0

    def __call__(self, name, size, total):
        if size > 0 and total > 0:
            self._seen += size
            pct = self._seen * 100.0 / total
            sys.stderr.write('\rUploading %s - %.2f%%' % (name, pct))
            if int(pct) >= 100:
                sys.stderr.write('\n')


class DottedProgress(object):
    def __init__(self, full_length):
        self.__full_length = full_length
        self.__dots = 0
        self.__seen = 0.0

    def __call__(self, name, size, total):
        if size > 0 and total > 0:
            if self.__seen == 0:
                sys.stderr.write('Uploading %s: ' % name)
            self.__seen += size
            dots = int(self.__seen * self.__full_length / total)
            while dots > self.__dots and self.__dots < self.__full_length:
                self.__dots += 1
                sys.stderr.write('.')
            if self.__dots == self.__full_length:
                sys.stderr.write('\n')



from .version import __version__

from .exceptions import (GenestackAuthenticationException, GenestackBaseException,
                                  GenestackConnectionFailure, GenestackException,
                                  GenestackResponseError, GenestackServerException,
                                  GenestackVersionException)
from .connection import Connection, Application
from .file_types import FileTypes
from .file_permissions import Permissions
from .metainfo_scalar_values import *
from .bio_meta_keys import BioMetaKeys
from .metainfo import Metainfo
from .data_importer import DataImporter
from .genome_query import GenomeQuery
from .utils import get_connection, get_user, make_connection_parser, validate_constant
from .file_filters import *
from .share_util import ShareUtil
from .files_util import FilesUtil, SortOrder, SpecialFolders
from .datasets_util import DatasetsUtil
from .groups_util import GroupsUtil
from .task_log_viewer import TaskLogViewer
from .cla import *
from .expression_navigator import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script loads, inits and shares dictionaries which locations are supplied in a separate file.
# Example file with a list with dictionaries can be found in example-data directory.
# Those dictionaries which are supplied from the local machine should have a full path to their
# location in 'url' parameter or they should be present in the same directory as the script.
# The same works for the json file with dictionaries paths.
# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/945094729/How+to+load+init+and+share+a+dictionary+on+a+remote+host+by+executing+a+script+on+your+machine+CUT+prepared+for+outer+clients
import json

from odm_sdk import (DataImporter, FilesUtil,
                              GenestackException, GenestackServerException,
                              Metainfo, ShareUtil, SpecialFolders)
from odm_sdk.utils import get_connection, make_connection_parser

from odm_sdk.scripts.utils import colored, GREEN, BLUE, RED


def load_dictionary(connection, data, parent_dictionary=None, replace=True,
                    reuse_old_version=False, metainfo=None):

    if metainfo:
        metainfo.add_string(Metainfo.DESCRIPTION, data.get('description'))
    else:
        metainfo = data.get('metainfo', Metainfo())
        metainfo.add_string(Metainfo.DESCRIPTION, data.get('description'))

    name = data['name']
    url = data['url']
    term_type = data.get('term_type')

    fu = FilesUtil(connection)
    parent = fu.get_folder(
        fu.get_special_folder(SpecialFolders.CREATED),
        'Data samples',
        'Dictionaries',
        create=True)

    di = DataImporter(connection)
    old_dictionary_version = fu.find_file_by_name(name, parent=parent)
    if old_dictionary_version:
        if replace:
            print('Old version of dictionary %s / %s is removed'
                  % (colored(old_dictionary_version, GREEN), colored(name, BLUE)))
            fu.mark_obsolete(old_dictionary_version)
            fu.unlink_file(old_dictionary_version, parent)
        else:
            if reuse_old_version:
                print('Dictionary %s / %s already exists and will be reused'
                      % (colored(old_dictionary_version, GREEN), colored(name, BLUE)))
                return old_dictionary_version
            raise GenestackException(
                "Dictionary %s / %s already exists, set replace=True to replace it"
                % (colored(old_dictionary_version, GREEN), colored(name, BLUE)))

    accession = di.create_dictionary(
        parent=parent,
        name=name,
        url=url,
        term_type=term_type,
        metainfo=metainfo,
        parent_dictionary=parent_dictionary
        )
    print_result(accession, name, 'Dictionary')
    return accession


def print_result(accession, name, type_name, started=False):
    print('%s %s %s / %s' % (
        'Created and started' if started else 'Created',
        type_name,
        colored(accession, GREEN), colored(name, BLUE)
    ))


def initialization(connection, accessions):
    try:
        FilesUtil(connection).initialize(accessions)
        print('Initialization of %s dictionaries started.' % len(accessions))
    except GenestackServerException:
        print(colored("Created dictionaries have not been initialized. "
                      "Re-run this script as public@genestack.com "
                      "in order to share them with everyone", RED))


def sharing(connection, accessions):
    try:
        all_users_group_accession = 'GSG000001'
        ShareUtil(connection).share_files_for_view(accessions, all_users_group_accession,
                                                   'public')
    except GenestackServerException:
        print(colored("Created dictionaries have not been shared. "
                      "Re-run this script as public@genestack.com "
                      "in order to share them with everyone", RED))


def main():
    try:
        # load
        args = get_arguments()
        connection = get_connection(args)
        with open(args.file_with_dictionaries, 'r') as data_file:
            dictionaries = json.load(data_file)
        accessions = [load_dictionary(connection, data) for data in dictionaries]
        initialization(connection, accessions)
        sharing(connection, accessions)

    except Exception as e:
        print(colored(e, RED))


def get_arguments():
    parser = make_connection_parser()
    group = parser.add_argument_group('required arguments')
    group.add_argument('--file_with_dictionaries', metavar='<file_with_dictionaries>',
                       default="dictionaries.json",
                       help='dictionaries to load', required=True)
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    main()



READS_LOCATION = 'genestack.location:reads'
READS_LINK = 'genestack.url:reads'


class Key(object):
    SPACE = 'space'
    FORMAT = 'format'
    TYPE = 'type'


class Space(object):
    BASESPACE = 'basespace'
    COLORSPACE = 'colorspace'


class Format(object):
    PHRED33 = 'phred33'
    PHRED64 = 'phred64'
    FASTA_QUAL = 'fasta-qual'
    SRA = 'sra'
    SFF = 'sff'
    FAST5 = 'fast5'


class Type(object):
    SINGLE = 'single'
    PAIRED = 'paired'
    PAIRED_WITH_UNPAIRED = 'paired-with-unpaired'


def compose_format_map(space, file_format, file_type):
    return {Key.SPACE: space,
            Key.FORMAT: file_format,
            Key.TYPE: file_type}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script shares a study with a group.

# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/1262419987/How+to+share+a+study
from __future__ import print_function, unicode_literals

import re
from odm_sdk import GenestackServerException, GroupsUtil, ShareUtil
from odm_sdk.utils import make_connection_parser, get_connection

from odm_sdk.scripts.utils import colored, RED, GREEN


def share_study(share_params):
    connection = get_connection(share_params)
    study_accession = share_params.study_accession
    try:
        groups_util = GroupsUtil(connection)
        group_accession = groups_util.find_group_by_name(share_params.group_name)
        share_util = ShareUtil(connection)
        share_util.share_files_for_view(study_accession, group_accession)
        print(colored('Success', GREEN))
    except GenestackServerException as e:
        __handle_error_and_exit_with_code(e, study_accession)


def __handle_error_and_exit_with_code(e, study_accession):
    raw_to_custom_error_map = {
        r'File .* not found': f'Study with accession {study_accession} does not exist',
        r'Not enough user permissions to call method .*': 'Only the owner can share a study.'
    }

    for raw_err, custom_err in raw_to_custom_error_map.items():
        if re.search(raw_err, e.message):
            print(colored(custom_err, RED))
            exit(2)

    print(colored(e, RED))
    exit(1)


class ShareParams:
    # The class constructor is a subject to change, its parameters might be changed in the future.
    def __init__(
            self,
            host,
            study_accession,
            group_name,
            token=None,
            access_token=None,
            user=None,
            pwd=None,
            show_logs=False,
            debug=False
    ):
        self.host = host.rstrip('/')
        self.token = token
        self.access_token = access_token
        self.user = user
        self.pwd = pwd
        self.study_accession = study_accession
        self.group_name = group_name
        self.show_logs = show_logs
        self.debug = debug

    @classmethod
    def from_parsed_params(cls, args):
        return cls(
            host=args.host,
            study_accession=args.study_accession,
            group_name=args.group_name,
            token=args.token,
            access_token=args.access_token,
            user=args.user,
            pwd=args.pwd,
            show_logs=args.show_logs,
            debug=args.debug
        )


def main():
    parser = make_connection_parser()
    group = parser.add_argument_group('required arguments')
    group.add_argument('--study_accession', metavar='<study_accession>',
                       help='accession of study to share', required=True)
    group.add_argument('--group_name', metavar='<group_name>',
                       help='group name with which to share', required=True)
    args = parser.parse_args()
    share_study(ShareParams.from_parsed_params(args))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2011-2023 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
#

from odm_sdk import (DataImporter, FilesUtil,
                              SpecialFolders, get_connection, make_connection_parser)


def main():
    """
    Simple wrapper for uploading reference genome from public domain:  https://www.ensembl.org/index.html

    For the details of FASTA file naming conventions, please see the README at:
    ftp://ftp.ensembl.org/pub/release-91/fasta/homo_sapiens/dna/README
    """

    args = get_args()
    connection = get_connection(args)
    organism = args.species.replace("_", " ")

    # Generate annotation_url if not specified
    if args.annotation_url:
        annotation_url = args.annotation_url
    else:
        annotation_url = f"ftp://ftp.ensembl.org/pub/release-{args.release}/gtf/{args.species.lower()}/" \
                         f"{args.species}.{args.assembly}.{args.release}.gtf.gz"
    # TODO: Use logger in future
    print(f"Reference genome source URL:\n{annotation_url}")

    # Generate name if not specified
    if args.name:
        name = args.name
    else:
        name = f"{organism} reference genome {args.assembly}.{args.release}"

    # TODO: Use logger in future
    print(f"Reference genome name:\n{name}")

    # Create folder in GenestackFS
    fu = FilesUtil(connection)
    parent = fu.get_folder(fu.get_special_folder(SpecialFolders.CREATED),
                           "Data samples", "Reference genome",
                           create=True)

    # Make api-call
    accession = DataImporter(connection).create_reference_genome(
        parent,
        name=name,
        description=name,
        organism=organism,
        assembly=args.assembly,
        release=args.release,
        annotation_url=annotation_url
    )
    files_util = FilesUtil(connection)
    files_util.initialize([accession])


def get_args():

    parser = make_connection_parser()

    parser.add_argument("--name",
                        help="[Optional] Reference genome name to be set in ODM. "
                             "The name will be generated automatically when the parameter is not specified. ")

    parser.add_argument("--annotation-url",
                        help="[Optional] Direct url to a GTF or GFF annotation file. "
                        "The url will be generated automatically when the parameter is not specified, "
                        "e.g. https://ftp.ensembl.org/pub/release-109/gtf/homo_sapiens/Homo_sapiens.GRCh38.109.gtf.gz")

    # Specific group for required args
    group = parser.add_argument_group("required arguments")

    group.add_argument("--assembly", required=True,
                       help="The genome assembly version (excluding the patch number), "
                       "e.g. \"GRCh38\" or \"mm10\"")
    group.add_argument("--release", required=True,
                       help="The release version according to Ensembl, "
                       "e.g. \"109\"")
    group.add_argument("--species", required=True,
                       help="The systematic name of the species with underscores, "
                       "e.g., \"Homo_sapiens\" or \"Mus_musculus\"")

    return parser.parse_args()


if __name__ == "__main__":
    main()



import sys
import time

from odm_sdk import Application, GenestackException


class TaskLogViewer(Application):
    """
    A wrapper class for the Task Logs Viewer application. This application allows you to access the initialization
    logs of a file.
    """
    APPLICATION_ID = 'genestack/task-log-viewer'

    STDERR = 'stderr'
    STDOUT = 'stdout'
    MAX_CHUNK_SIZE = 128000

    def print_log(self, accession, log_type=None, follow=True, offset=0):
        """
        Print a file's latest task initialization logs to `stdout`.
        Raises an exception if the file is not found or has no associated initialization task.
        By default the output `stdout` log is shown. You can also view the `stderr` error log.
        ``follow=True`` will wait until initialization is finished.
        Incoming logs will be printed to the console.

        :param accession: file accession
        :param log_type: `stdout` or `stderr`
        :param follow: if enabled, wait and display new lines as they appear (similar to ``tail --follow``)
        :param offset: offset from which to start retrieving the logs. Set to `-1` if you want to start retrieving
          logs from the latest chunk.
        """
        if not log_type:
            log_type = self.STDOUT
        waiting_message_shown = True

        while True:
            log_chunk = self.invoke('getFileInitializationLog', accession, log_type, offset, self.MAX_CHUNK_SIZE)
            if not log_chunk:
                raise GenestackException('File %s not found or has no initialization task' % accession)

            if log_chunk['content'] is not None:
                sys.stdout.write(log_chunk['content'])
                sys.stdout.flush()
                offset += len(log_chunk['content'])
            else:
                # the current chunk is empty
                if offset == 0:
                    if log_chunk['isTerminal']:
                        print('No logs were produced.')
                    elif follow and not waiting_message_shown:
                        print('No logs produced yet...')
                        waiting_message_shown = True

                if log_chunk['isTerminal'] or not follow:
                    break

                if follow:
                    time.sleep(0.5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script deletes files using wipeStudy method.
# Provide an accession of a study/template which needs to be deleted.

from __future__ import print_function, unicode_literals

import re

from odm_sdk import GenestackServerException
from odm_sdk.utils import make_connection_parser, get_connection

from odm_sdk.scripts.utils import colored, GREEN, RED


def main():
    parser = make_connection_parser()
    group = parser.add_argument_group('required arguments')
    group.add_argument('--accession', metavar='<accession>',
                       help='accession of a study/template to delete', required=True)
    args = parser.parse_args()
    connection = get_connection(args)

    accession = args.accession
    try:
        connection.application('genestack/study-metainfo-editor').invoke('wipeStudy', accession)
        print(colored("Success", GREEN))
    except GenestackServerException as e:
        p = re.compile('File .* not found')
        result = p.search(e.message)
        if result is not None:
            print(colored("Study/template with accession %s does not exist" % accession,
                          RED))
        else:
            print(colored(e, RED))


if __name__ == "__main__":
    main()



#
# Copyright (c) 2011-2023 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
#



#!/usr/bin/env python3

#
# Copyright (c) 2011-2021 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
#

import sys
import time
from odm_sdk import get_connection

TASKMANAGER_ID = "genestack/taskmanager"
TIMEOUT_IN_SECONDS = 3 * 60 * 60
SLEEP_TIMEOUT_IN_SECONDS = 30

# The number of running tasks must be 0 several times in row to get a success
NO_RUNNING_TASK_ITERATION_LIMIT = 3


def main():
    connection = get_connection()
    task_app = connection.application(TASKMANAGER_ID)

    no_running_task_check_iteration = 1
    start_time = time.time()

    while time.time() - start_time < TIMEOUT_IN_SECONDS:
        time.sleep(SLEEP_TIMEOUT_IN_SECONDS)
        tasks = task_app.invoke('getRecentTaskCountsByStatus', 20)
        print(
            f'Check for no running tasks iteration: {no_running_task_check_iteration} out of {NO_RUNNING_TASK_ITERATION_LIMIT}\n'
            f'Running tasks: {tasks["running"]}\n'
            f'Queued tasks: {tasks["queued"]}'
        )

        if tasks["running"] == 0:
            no_running_task_check_iteration += 1
        else:
            no_running_task_check_iteration = 1
            continue

        if no_running_task_check_iteration > NO_RUNNING_TASK_ITERATION_LIMIT:
            print("All tasks complete.")
            sys.exit(0)

    print("Waiting timeout exceeded.")
    sys.exit(1)


if __name__ == "__main__":
    main()



from getpass import getpass
from urllib.parse import urlsplit

import requests

from odm_sdk import Connection, GenestackAuthenticationException, GenestackException
from odm_sdk.utils import isatty, interactive_select


def _get_server_url(host):
    split_host = urlsplit(host)
    has_scheme = bool(split_host.scheme)

    # default to HTTPS if not explicitly defined
    url_stub = host if has_scheme else 'https://%s' % host
    url_stub = url_stub.rstrip('/')
    # trust provided path if it ends with '/frontend'
    if url_stub.split('/')[-1] == 'frontend':
        return '/'.join([url_stub, 'endpoint'])
    # return '.../frontend/endpoint' if '.../frontend/health' works
    if requests.get('%s/frontend/health' % url_stub).ok:
        return '%s/frontend/endpoint' % url_stub

    raise GenestackAuthenticationException(
        "Could not connect to host '{}', check if it is defined correctly"
        "".format(host))


def _read_non_canonical(msg):
    """
    An enhanced input method designed to read user inputs that may exceed the standard terminal input buffer size,
    typically limited to 1024 characters. It is required for interactively reading lengthy inputs such as raw access
    tokens directly from the user.
    The function attempts to switch the terminal to non-canonical mode, this mode is only supported on Unix-like
    systems where the termios module is available.
    :param msg: The prompt message displayed to the user.
    :return: The user input, stripped of leading and trailing whitespace.
    """
    try:
        import termios, tty, atexit, sys
        termios.tcgetattr, termios.tcsetattr
    except (ImportError, AttributeError):
        # fallback to standard input on systems that don't support termios
        return input(msg).strip()
    fd = sys.stdin.fileno()
    original_attrs = termios.tcgetattr(fd)
    # ensure that terminal settings are restored to their original state on program exit
    atexit.register(lambda: termios.tcsetattr(fd, termios.TCSADRAIN, original_attrs))

    # temporarily switch to non-canonical mode modifying local modes
    new_attrs = termios.tcgetattr(fd)
    new_attrs[3] = new_attrs[3] & ~termios.ICANON
    termios.tcsetattr(fd, termios.TCSADRAIN, new_attrs)
    try:
        return input(msg).strip()
    finally:
        # restore the original settings
        termios.tcsetattr(fd, termios.TCSADRAIN, original_attrs)


class User(object):
    """
    Class encapsulating all user info required for authentication.

    That includes:
     - user alias
     - server URL (or is it hostname?)
     - token *or* email/password pair
    """
    def __init__(self, email=None, alias=None, host=None, password=None, token=None, access_token=None):
        """
        All fields are optional.
        If ``alias`` is None it will be the same as ``email``.

        If you login interactively, no ``email`` or ``password`` is required.
        The alias is used to find the matching user in :py:func:`~odm_sdk.get_user`

        :param email: email
        :type email: str
        :param alias: alias
        :type alias: str
        :param host: host
        :type host: str
        :param password: password
        :type password: str
        """
        self.host = host
        self.email = email
        self.password = password  # TODO make property
        self.alias = alias or email
        self.token = token
        self.access_token = access_token

    def get_connection(self, interactive=True, debug=False, show_logs=False):
        """
        Return a logged-in connection for current user.
        If ``interactive`` is ``True`` and the password or email are unknown,
        they will be asked in interactive mode.

        :param interactive: ask email and/or password interactively.
        :type interactive: bool
        :param debug: print stack trace in case of exception
        :type debug: bool
        :param show_logs: print application logs (received from server)
        :type show_logs: bool
        :return: logged connection
        :rtype: odm_sdk.Connection
        """
        connection = Connection(_get_server_url(self.host), debug=debug, show_logs=show_logs)
        if self.token:
            connection.login_by_token(self.token)
        elif self.access_token:
            connection.login_by_access_token(self.access_token)
        elif self.email and self.password:
            connection.login(self.email, self.password)
        elif interactive:
            self.__interactive_login(connection)
        #else:
        #    raise GenestackException('Not enough user data to login')
        return connection

    def __repr__(self):
        return "User('%s', alias='%s', host='%s', password='%s', token='%s')" % (
            self.email, self.alias, self.host, self.password and '*****', self.token and '*****')

    def __interactive_login(self, connection):
        if not isatty():
            raise GenestackException("Interactive login is not possible")

        email = self.email
        message = 'Connecting to %s' % self.host

        login_by_token = 'by token'
        login_by_access_token = 'by access token'
        login_by_email = 'by email and password'
        login_anonymously = 'anonymously'

        choice = interactive_select([login_by_token, login_by_access_token, login_by_email, login_anonymously],
                                    'How do you want to login')

        if choice == login_anonymously:
            return

        while True:
            if message:
                print(message)

            if choice == login_by_email:
                input_message = 'e-mail [%s]: ' % email if email and '@' in email else 'e-mail: '
                email = input(input_message).strip() or email

                password = getpass('password for %s: ' % email)
                try:
                    connection.login(email, password)
                    self.email = email
                    self.password = password
                    return
                except GenestackAuthenticationException:
                    message = ('Your username and password have been rejected by %s, '
                               'please try again' % self.host)
            elif choice == login_by_access_token:
                access_token = _read_non_canonical('access token or environment variable with its value: ')
                try:
                    connection.login_by_access_token(access_token)
                    self.access_token = access_token
                    return
                except GenestackAuthenticationException:
                    message = 'Your access token has been rejected by %s, please try again' % self.host
            else:
                token = getpass('token: ')
                try:
                    connection.login_by_token(token)
                    self.token = token
                    return
                except GenestackAuthenticationException:
                    message = 'Your token has been rejected by %s, please try again' % self.host

    def __eq__(self, other):
        return (
            isinstance(other, User) and
            self.alias == other.alias and
            self.password == other.password and
            self.host == other.host and
            self.email == other.email
        )

    def __ne__(self, other):
        return not self.__eq__(other)



from odm_sdk import Application


class GroupsUtil(Application):
    APPLICATION_ID = 'genestack/groupsUtil'

    def find_group_by_name(self, name):
        """
        Finds group with specified name.
        If there are no groups or more than one group with this name, an exception is thrown.

        :param name: group name
        :type name: str
        :return: group accession
        :rtype: str
        """
        return self.invoke(
            'findGroupByName', name
        )



import sys
from odm_sdk import (BooleanValue, DateTimeValue, DecimalValue, ExternalLink,
                              FileReference, GenestackException, IntegerValue, MemorySizeValue,
                              MetainfoScalarValue, Organization, Person,
                              Publication, StringValue)


class Metainfo(dict):
    """
    A Python representation of metainfo objects.
    """
    NAME = 'genestack:name'
    DESCRIPTION = 'genestack:description'
    ACCESSION = 'genestack:accession'
    ORGANIZATION = 'genestack:organization'
    CONTACT_PERSON = 'genestack:contactPerson'
    EXTERNAL_LINK = 'genestack:links'
    CREATION_DATE = 'genestack:dateCreated'
    PARENT_DICTIONARY = 'genestack.dictionary:parent'
    SOURCE_DATA = 'genestack:sourceData'
    DATA_TYPE = 'genestack:dataType'
    LAST_UPDATE_DATE = 'genestack:file.last-update'

    # Metainfo key for links from data files to their samples.
    SAMPLE_LINK = "genestack:sampleLink"
    # Metainfo key for links from the dataset with data files to the according study.
    STUDY_LINK = "genestack:studyLink"

    YEAR = 'YEAR'
    MONTH = 'MONTH'
    WEEK = 'WEEK'
    DAY = 'DAY'
    HOUR = 'HOUR'
    MINUTE = 'MINUTE'
    SECOND = 'SECOND'
    MILLISECOND = 'MILLISECOND'

    CELSIUS = 'CELSIUS'
    KELVIN = 'KELVIN'
    FAHRENHEIT = 'FAHRENHEIT'

    def _add_value(self, key, value, type):
        self.setdefault(key, []).append({'type': type, 'value': value})

    @staticmethod
    def _create_dict_with_type(type):
        return {'type': type}

    def add_value(self, key, value):
        """
        Add a scalar value to a metainfo key.
        If adding to an existing key, the value will be appended to the list of existing values.
        :param key: key
        :type key: str
        :param value: value
        :type value: MetainfoScalarValue
        :rtype None:
        """
        if not isinstance(value, MetainfoScalarValue):
            raise GenestackException("Value is not an instance of `MetainfoScalarValue`")
        self.setdefault(key, []).append(value)

    def add_string(self, key, value):
        """
        Add a string value.

        :param key: key
        :type key: str
        :param value: string value
        :type value: str
        :rtype: None
        """
        self.add_value(key, StringValue(value))

    def add_boolean(self, key, value):
        """
        Add a boolean value.

        :param key: key
        :type key: str
        :param value: boolean value
        :type value: bool
        :rtype: None
        """
        self.add_value(key, BooleanValue(value))

    def add_integer(self, key, value):
        """
        Add an integer value.

        :param key: key
        :type key: str
        :param value: integer value
        :type value: int
        :rtype: None
        """
        self.add_value(key, IntegerValue(value))

    def add_memory_size(self, key, value):
        """
        Add a memory size in bytes.

        :param key: key
        :type key: str
        :param value: integer value
        :type value: int
        :rtype: None
        """
        self.add_value(key, MemorySizeValue(value))

    def add_decimal(self, key, value):
        """
        Add a decimal value.

        :param key: key
        :type key: str
        :param value: integer value
        :type value: float | str
        :rtype: None
        """
        self.add_value(key, DecimalValue(value))

    def add_external_link(self, key, url, text=None, fmt=None):
        """
        Add an external link. The URL should point to a valid source file.
        The source should be either a publicly available file on the web, or a local file.
        Local files will be uploaded if imported with :py:class:`~odm_sdk.DataImporter`

        :param key: key
        :type key: str
        :param text: URL text for display purposes
        :type text: str
        :param fmt: format for an unaligned reads link
        :type fmt: dict
        :rtype: None
        """
        self.add_value(key, ExternalLink(url, text, fmt))

    def add_person(self, key, name, phone=None, email=None):
        """
        Add a person. The name is required, and all other fields are optional.
        All fields will be visible to anyone who has access to this metainfo object.

        :param key: key
        :type key: str
        :param name: full name
        :type name: str
        :param phone: phone number
        :type phone: str
        :param email: contact email
        :type email: str
        :rtype: None

        **Deprecated** since *0.32.0*, use compound metainfo keys instead
        """
        self._print_metainfo_type_deprecation_message('add_person')
        self.add_value(key, Person(name, phone, email))

    def add_publication(self, key, title, authors, journal_name,
                        issue_date, identifiers=None, issue_number=None, pages=None):
        """
        Add a publication.
        All fields will be visible to anyone who has access to this metainfo object.

        :param key:
        :type key: str
        :param title: publication title
        :type title: str
        :param identifiers: publication identifiers
        :type identifiers: dict
        :param authors: publication authors
        :type authors: str
        :param journal_name: name of the journal containing this publication
        :type journal_name: str
        :param issue_date: journal issue date
        :type issue_date: str
        :param issue_number: journal issue number
        :type issue_number: str
        :param pages: pages in the journal issue
        :type pages: str
        :rtype: None

        **Deprecated** since *0.32.0*, use compound metainfo keys instead
        """
        self._print_metainfo_type_deprecation_message('add_publication')
        self.add_value(key, Publication(title, authors, journal_name, issue_date,
                                        identifiers, issue_number, pages))

    def add_organization(self, key, name, department=None, country=None, city=None, street=None,
                         postal_code=None, state=None, phone=None, email=None, url=None):
        """
        Add an organization. The name is required, and all other fields are optional.
        All fields will be visible to anyone who has access to this metainfo object.

        :param key: key
        :type key: str
        :param name: name
        :type name: str
        :param department: department
        :type department: str
        :param country: country
        :type country: str
        :param city: city
        :type city: str
        :param street: street
        :type street: str
        :param postal_code: postal/zip code
        :type postal_code: str
        :param state: state
        :type state: str
        :param phone: phone
        :type phone: str
        :param email: email
        :type email: str
        :param url: organisation web page
        :type url: str
        :rtype: None

        **Deprecated** since *0.32.0*, use compound metainfo keys instead
        """
        self._print_metainfo_type_deprecation_message('add_organization')
        self.add_value(key, Organization(name, department, country, city, street,
                                         postal_code, state, phone, email, url))

    def add_time(self, key, value, unit):
        """
        Add a time value (like an age, or the duration of an experiment for example).

        The value can be any number, supplied with a unit from a controlled vocabulary.

        The time unit should be one of the following:
            :py:attr:`~odm_sdk.Metainfo.YEAR`,
            :py:attr:`~odm_sdk.Metainfo.MONTH`,
            :py:attr:`~odm_sdk.Metainfo.WEEK`,
            :py:attr:`~odm_sdk.Metainfo.DAY`,
            :py:attr:`~odm_sdk.Metainfo.HOUR`,
            :py:attr:`~odm_sdk.Metainfo.MINUTE`,
            :py:attr:`~odm_sdk.Metainfo.SECOND`,
            :py:attr:`~odm_sdk.Metainfo.MILLISECOND`

        :param key: key
        :type key: str
        :param: number of units as float
        :type value: float | str
        :param unit: unit
        :type unit: str
        :rtype: None

        **Deprecated** since *0.32.0*, use compound metainfo keys instead
        """
        self._print_metainfo_type_deprecation_message('add_time')
        result = Metainfo._create_dict_with_type('time')
        result['value'] = value
        result['unit'] = unit.upper()
        self.setdefault(key, []).append(result)

    def add_temperature(self, key, value, unit):
        """
        Add a temperature value.
        The value can be any number, supplied with a unit from a controlled vocabulary.


        The temperature unit should be one of the following:
            :py:attr:`~odm_sdk.Metainfo.CELSIUS`,
            :py:attr:`~odm_sdk.Metainfo.KELVIN`,
            :py:attr:`~odm_sdk.Metainfo.FAHRENHEIT`,

        :param key: key
        :type key: str
        :param value: number of units as float
        :type value: float | str
        :param unit: unit
        :type unit: str
        :rtype: None

        **Deprecated** since *0.32.0*, use compound metainfo keys instead
        """
        self._print_metainfo_type_deprecation_message('add_temperature')
        result = Metainfo._create_dict_with_type('temperature')
        result['value'] = value
        result['unit'] = unit.upper()
        self.setdefault(key, []).append(result)

    def add_file_reference(self, key, accession):
        """
        Add a reference to another Genestack file.

        :param key: key
        :type key: str
        :param accession: accession of the file to reference
        :type accession: str
        :rtype: None
        """
        self.add_value(key, FileReference(accession))

    def add_date_time(self, key, time):
        """
        Add a date.
        The time parameter can be passed in one of the following formats:

         - :py:class:`datetime.datetime`
         - :py:class:`datetime.date`
         - :py:class:`str` in format: ``'%Y-%m-%d %H:%M:%S'`` or ``'%Y-%m-%d'``
         - number of seconds since the epoch as a floating point number

        :param key: key
        :type key: str
        :param time: time value
        :rtype: None
        """
        self.add_value(key, DateTimeValue(time))

    @classmethod
    def parse_metainfo_from_dict(cls, source_dict):
        """
        Parse a Java map representing a metainfo object and create a Python Client Metainfo.
        :param source_dict: Java map
        :type source_dict: dict
        :rtype: Metainfo
        """
        output = Metainfo()
        for key in source_dict:
            for scalar_value in map(cls._parse_scalar_value, source_dict[key]):
                output.add_value(key, scalar_value)
        return output

    @staticmethod
    def _parse_scalar_value(source_dict):
        java_type = source_dict['type']
        if java_type == BooleanValue._TYPE:
            return BooleanValue(source_dict['value'])
        elif java_type == DateTimeValue._TYPE:
            return DateTimeValue(source_dict['date'])
        elif java_type == ExternalLink._TYPE:
            return ExternalLink(
                source_dict['url'], source_dict['text'],
                source_dict.get('format')
            )
        elif java_type == IntegerValue._TYPE:
            return IntegerValue(source_dict['value'])
        elif java_type == DecimalValue._TYPE:
            return DecimalValue(source_dict['value'])
        elif java_type == MemorySizeValue._TYPE:
            return MemorySizeValue(source_dict['value'])
        elif java_type == Organization._TYPE:
            return Organization(
                source_dict['name'], source_dict['department'],
                source_dict['street'], source_dict['city'],
                source_dict['state'], source_dict['postalCode'],
                source_dict['country'], source_dict['email'],
                source_dict['phone'], source_dict['url']
            )
        elif java_type == Person._TYPE:
            return Person(source_dict['name'], source_dict['phone'], source_dict['email'])
        elif java_type == Publication._TYPE:
            return Publication(
                source_dict['title'], source_dict['authors'],
                source_dict['journalName'], source_dict['issueDate'],
                source_dict['identifiers'], source_dict['issueNumber'],
                source_dict['pages']
            )
        elif java_type == StringValue._TYPE:
            return StringValue(source_dict['value'])
        elif java_type == FileReference._TYPE:
            return FileReference(source_dict['accession'])
        else:
            # this is safer than raising an exception, since new metainfo types
            # can be added in Java (and some deprecated ones like physical values are not handled here)
            return MetainfoScalarValue(source_dict)


    @staticmethod
    def _print_metainfo_type_deprecation_message(method_name):
        sys.stderr.write(
            "Method `%s` is deprecated. Use compound metainfo keys to store complex values "
            "(e.g. two separate keys: 'myKey/value' and 'myKey/unit').\n" % method_name
        )



import collections

from odm_sdk import Application

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


class ShareUtil(Application):
    """
    Application that acts as a facade for sharing-related operations.
    """
    APPLICATION_ID = 'genestack/shareutils'
    SHARE_FOLDER_LIMIT = 100

    class Permissions(object):
        """
        Supported permission values that can be used in :meth:`ShareUtil.share_files` and
        :meth:`ShareUtil.share_folder` methods.

        .. attribute:: VIEW

            Allows finding files via search and reading files' content

        .. attribute:: EDIT

            Allows finding files via search, reading files' content and modifying files' metainfo

        .. attribute:: SHARE

            Allows finding files via search, reading files' content and sharing them with other
            groups. This permissions type only allows sharing by group members from the same
            organization as the file owner. When sharing, non-owners are only allowed to set
            permissions that are the same or narrower that they currently have for the given file.

        """
        VIEW = "VIEW"
        EDIT = "EDIT"
        SHARE = "SHARE"

    def share_files_for_view(self, file_accessions, group_accession, destination_folder=None):
        """
        Share files with viewing permissions. Viewing permissions include finding the shared files
        and running tasks that access their content.

        This method is equivalent to calling :meth:`share_files` method with
        :attr:`ShareUtil.Permissions.VIEW` permission.

        :param file_accessions: accession or an iterable of accessions of files to be shared
        :type file_accessions: str | collections.Iterable[str]
        :param group_accession: accession of the group to share the files with
        :type group_accession: str
        :param destination_folder: accession of the folder to link shared files into. Typically
                                   this parameter should be used for linking files into group
                                   folders, which is currently impossible to do using the
                                   :meth:`FilesUtil.link_file` method. No links will be created if
                                   this parameter is equal to `None`.
        :type destination_folder: str
        """
        self.share_files(
            file_accessions, group_accession, [ShareUtil.Permissions.VIEW], destination_folder
        )

    def share_files_for_edit(self, file_accessions, group_accession, destination_folder=None):
        """
        Share files with editing permissions. Editing permissions include viewing permissions and
        also allow modifying metainfo and linking/unlinking files (only applicable to containers
        and datasets).

        This method is equivalent to calling :meth:`share_files` method with
        :attr:`ShareUtil.Permissions.EDIT` permission.

        :param file_accessions: accession or an iterable of accessions of files to be shared
        :type file_accessions: str | collections.Iterable[str]
        :param group_accession: accession of the group to share the files with
        :type group_accession: str
        :param destination_folder: accession of the folder to link shared files into. Typically
               this parameter should be used for linking files into group folders, which is
               currently impossible to do using the :meth:`FilesUtil.link_file` method. No links
               will be created if this parameter is equal to `None`.
        :type destination_folder: str
        """
        self.share_files(
            file_accessions, group_accession, [ShareUtil.Permissions.EDIT], destination_folder
        )

    def share_files(self, file_accessions, group_accession, permissions, destination_folder=None):
        """
        Share files with the given permissions with the given groups. Available permission values
        are listed in the :class:`ShareUtil.Permissions` class.

        :param file_accessions: accession or an iterable of accessions of files to be shared
        :type file_accessions: str | collections.Iterable[str]
        :param group_accession: accession of the group to share the files with
        :type group_accession: str
        :param permissions: permissions that should be assigned to the provided files. Must consist
               of :class:`ShareUtil.Permissions` values
        :type permissions: str | collections.Iterable[str]
        :param destination_folder: accession of the folder to link shared files into. Typically
               this parameter should be used for linking files into group folders, which is
               currently impossible to do using the :meth:`FilesUtil.link_file` method. No links
               will be created if this parameter is equal to `None`
        :type destination_folder: str

        :raises GenestackServerException: if some of the given files cannot be shared by the current
                user (i.e. he doesn't own them or doesn't have the
                :attr:`ShareUtil.Permissions.SHARE` permission).
        """
        permissions = self.__to_list(permissions)
        self.__share(
            file_accessions, group_accession, destination_folder, 'shareFiles', permissions
        )

    def safe_share_files(
            self, file_accessions, group_accession, permissions, destination_folder=None
    ):
        """
        Same as :meth:`share_files` but does not throw an exception in case some of the given files
        cannot be shared (i.e. the current user doesn't own them or doesn't have the
        :attr:`ShareUtil.Permissions.SHARE` permission).

        :param file_accessions: accession or an iterable of accessions of files to be shared
        :type file_accessions: str | collections.Iterable[str]
        :param group_accession: accession of the group to share the files with
        :type group_accession: str
        :param permissions: permissions that should be assigned to the provided files. Must consist
               of :class:`ShareUtil.Permissions` values
        :type permissions: str | collections.Iterable[str]
        :param destination_folder: accession of the folder to link shared files into. Typically
               this parameter should be used for linking files into group folders, which is
               currently impossible to do using the :meth:`FilesUtil.link_file` method. No links
               will be created if this parameter is equal to `None`
        :type destination_folder: str
        """
        permissions = self.__to_list(permissions)
        self.__share(
            file_accessions, group_accession, destination_folder, 'safeShareFiles', permissions
        )

    def __share(
            self, file_accessions, group_accession, destination_folder, method, *args
    ):
        file_accessions = self.__to_list(file_accessions)
        self.invoke(method, file_accessions, [group_accession], *args)
        if destination_folder is not None:
            self.invoke('linkFiles', file_accessions, destination_folder, group_accession)

    def share_folder(self, folder_accession, group_accession, permissions, destination_folder=None):
        """
        Recursively share the given folder, its subfolders and files inside them. Files that
        cannot be shared by the current user will be skipped.

        This method is useful for sharing folders with a lot of files because calling
        :meth:`share_files` may result in a timeout. This method shares files in chunks and may take
        significant time to complete.

        :param folder_accession: accession of the folder
        :type folder_accession: str
        :param group_accession: accession of the group to share the files with
        :type group_accession: str
        :param permissions: permissions that should be assigned to the provided files. Must consist
               of :class:`ShareUtil.Permissions` values
        :type permissions: str | collections.Iterable[str]
        :param destination_folder: accession of the folder to link shared files into. Typically
               this parameter should be used for linking files into group folders, which is
               currently impossible to do using the :meth:`FilesUtil.link_file` method. No links
               will be created if this parameter is equal to `None`
        :type destination_folder: str
        """

        # This method makes repeated calls to server, each method call
        # shares chunk of files in the given folder that aren't shared yet with the provided group
        # or ``WORLD``.
        # Due to the indexing lag, same files can be included in different calls. See the
        # `ShareUtilsApplication.shareChunkInFolder` javadoc for details.
        from time import sleep

        permissions = self.__to_list(permissions)

        self.share_files(folder_accession, group_accession, permissions, destination_folder)

        delay_seconds = 1  # delay between attempts

        offset = 0
        while True:
            count = self.invoke(
                'shareChunkInFolder',
                folder_accession, group_accession, permissions, offset, self.SHARE_FOLDER_LIMIT
            )
            if count == 0 and offset == 0:
                return
            if count < self.SHARE_FOLDER_LIMIT:
                sleep(delay_seconds)
                offset = 0
            else:
                offset += self.SHARE_FOLDER_LIMIT

    @staticmethod
    def __to_list(args):
        if isinstance(args, list):
            return args
        is_iterable = isinstance(args, collectionsAbc.Iterable)
        if is_iterable and not isinstance(args, str):
            return list(args)
        else:
            return [args]

    def get_available_sharing_groups(self):
        """
        Find groups that the current user can share files with, which means that he is either
        a sharing user or an administrator of these groups.

        :return: dictionary in format 'group accession' -> 'group name'
        :rtype: dict
        """
        return self.invoke('getGroupsToShare')



class FileTypes(object):

    _JAVA_PKG = "com.genestack.api.files."
    _JAVA_BIO_PKG = "com.genestack.bio.files."
    _JAVA_BTB_PKG = "com.genestack.api.files.btb."

    # Core file types
    FILE = _JAVA_PKG + "IFile"

    DATASET = _JAVA_PKG + "IDataset"

    AUXILIARY_FILE = _JAVA_PKG + "IAuxiliaryFile"
    CONTAINER = _JAVA_PKG + 'IContainerFile'
    FOLDER = _JAVA_PKG + "IFolder"
    SEARCH_FOLDER = _JAVA_PKG + "ISearchFolder"
    INDEX_FILE = _JAVA_PKG + "IIndexFile"

    APPLICATION_PAGE_FILE = _JAVA_PKG + "IApplicationPageFile"
    DICTIONARY_FILE = _JAVA_PKG + "IDictionaryFile"
    PREFERENCES_FILE = _JAVA_PKG + "IPreferencesFile"
    RAW_FILE = _JAVA_PKG + "IRawFile"
    REPORT_FILE = _JAVA_PKG + "IReportFile"
    SAMPLE = _JAVA_PKG + "ISample"

    # Biological file types
    REFERENCE_GENOME = _JAVA_BIO_PKG + "IReferenceGenome"
    VARIATION_FILE = _JAVA_BIO_PKG + "IVariationFile"
    EXTERNAL_DATABASE = _JAVA_BIO_PKG + "IExternalDataBase"

    CODON_TABLE = _JAVA_BIO_PKG + "ICodonTable"
    GENOME_ANNOTATIONS = _JAVA_BIO_PKG + "IGenomeAnnotations"
    HT_SEQ_COUNTS = _JAVA_BIO_PKG + "IHTSeqCounts"
    UNALIGNED_READS = _JAVA_BIO_PKG + "IUnalignedReads"
    MICROARRAY_DATA = _JAVA_BIO_PKG + "IMicroarrayData"
    ALIGNED_READS = _JAVA_BIO_PKG + "IAlignedReads"

    DIFFERENTIAL_EXPRESSION_FILE = _JAVA_BIO_PKG + "differentialExpression.IDifferentialExpressionFile"

    GENOME_BED_DATA = _JAVA_BIO_PKG + "IGenomeBEDData"
    GENOME_WIGGLE_DATA = _JAVA_BIO_PKG + "IGenomeWiggleData"

    FEATURE_LIST = _JAVA_BIO_PKG + "IFeatureList"
    GENE_EXPRESSION_SIGNATURE = _JAVA_BIO_PKG + "IGeneExpressionSignature"
    EXPRESSION_LEVELS = _JAVA_BIO_PKG + "IExpressionLevels"

    BTB_DOCUMENT = _JAVA_BTB_PKG + "IBTBDocumentFile"



class GenomeQuery(object):
    """
    Class describing a genome query.
    """
    _DELIMITER = '|'

    class SortingOrder(object):
        BY_FDR = 'ByPValue'
        BY_LOG_FOLD_CHANGE = 'ByLogFoldChange'
        BY_LOG_COUNTS = 'ByLogCountsPerMillion'

    class Filter(object):
        MAX_FDR = 'maximumFDR'
        MIN_LOG_FOLD_CHANGE = 'minimumLogFoldChange'
        MIN_LOG_COUNTS = 'minimumLogCountsPerMillion'
        REGULATION = 'regulation'

    class Regulation(object):
        UP = 'up'
        DOWN = 'down'

    def __init__(self):
        """
        Create a new genome query.
        The default parameters for a query are:

            - offset = 0
            - limit = 5000
            - no filters
            - search across all contrasts
            - sorting by increasing FDR

        :rtype: GenomeQuery
        """
        self._map = {'filter': {}}

    def set_feature_ids(self, features):
        features = features if isinstance(features, list) else [features]
        self._map['featureId'] = self._DELIMITER.join(features)
        return self

    def set_offset(self, offset):
        self._map['offset'] = offset
        return self

    def set_limit(self, limit):
        """
        Set maximum number of entries to retrieve **per contrast**.

        :param limit:
        :return:
        """
        self._map['limit'] = limit
        return self

    def set_order_ascending(self, ascending):
        self._map['ascending'] = ascending
        return self

    def set_contrasts(self, contrasts):
        contrasts = contrasts if isinstance(contrasts, list) else [contrasts]
        self._map['contrastLevel'] = self._DELIMITER.join(contrasts)
        return self

    def set_sorting_order(self, order):
        self._map['sortingOrder'] = order
        return self

    def add_filter(self, key, value):
        self._map['filter'][key] = str(value)

    def get_map(self):
        return self._map.copy()



from odm_sdk import Application, FilesUtil, SpecialFolders, Metainfo, GenestackException


class DatasetsUtil(Application):
    APPLICATION_ID = 'genestack/datasetsUtil'

    BATCH_SIZE = 100

    def create_dataset(self, name, dataset_type, children, parent=None, dataset_metainfo=None):
        """
        Create a dataset.

        :param name: name of the dataset
        :type name: str
        :param dataset_type: type of the dataset (children files interface name, must extend IDataFile)
        :type dataset_type: str
        :param children: list of children accessions
        :type children: list[str]
        :param parent: folder for the new dataset, 'My datasets' if not specified
        :type parent: str
        :param dataset_metainfo: metainfo of the created dataset
        :type dataset_metainfo: Metainfo

        :return: dataset accession
        :rtype: str
        """
        if parent is None:
            parent = self.__get_mydatasets_folder()

        dataset_metainfo = self._fill_dataset_metainfo(dataset_metainfo, name)

        return self.invoke('createDataset', parent, dataset_type, dataset_metainfo, children)

    def create_empty_dataset(self, name, dataset_type, parent=None, dataset_metainfo=None):
        """
        Create an empty dataset.

        :param name: name of the dataset
        :type name: str
        :param dataset_type: type of the dataset (children files interface name, must extend IDataFile)
        :type dataset_type: str
        :param parent: folder for the new dataset, 'My datasets' if not specified
        :type parent: str
        :param dataset_metainfo: metainfo of the created dataset
        :type dataset_metainfo: Metainfo

        :return: dataset accession
        :rtype: str
        """

        if parent is None:
            parent = self.__get_mydatasets_folder()

        dataset_metainfo = self._fill_dataset_metainfo(dataset_metainfo, name)

        return self.invoke('createEmptyDataset', parent, dataset_type, dataset_metainfo)

    @staticmethod
    def _fill_dataset_metainfo(dataset_metainfo, name):
        dataset_metainfo = dataset_metainfo or Metainfo()
        if Metainfo.NAME in dataset_metainfo:
            raise GenestackException(
                'Provided metainfo must not have "%s" field set' % Metainfo.NAME
            )

        dataset_metainfo.add_string(Metainfo.NAME, name)
        return dataset_metainfo

    def get_dataset_size(self, accession):
        """
        Get number of files in dataset.

        :param accession: dataset accession
        :type accession: str

        :return: number of files in dataset
        :rtype: int
        """
        return self.invoke('getDatasetSize', accession)

    def get_dataset_children(self, accession):
        """
        Return generator over children accessions of the provided dataset.

        :param accession: dataset accession
        :type accession: str
        :return: generator over dataset's children accessions
        """
        count = 0
        while True:
            children = self.invoke('getDatasetChildren', accession, count, self.BATCH_SIZE)
            for child in children:
                yield child

            count += len(children)
            if len(children) < self.BATCH_SIZE:
                break

    def create_subset(self, accession, children, parent=None):
        """
        Create a subset from dataset's children.

        :param accession: dataset accession
        :type accession: str
        :param children: list of children accessions to create a subset
        :type children: list[str]
        :param parent: folder for the new dataset, 'My datasets' if not specified
        :type parent: str

        :return: accession of the created subset
        :rtype: str
        """
        if parent is None:
            parent = self.__get_mydatasets_folder()

        return self.invoke('createSubset', parent, accession, children)

    def add_dataset_children(self, accession, children):
        """
        Add new files to a dataset.

        :param accession: dataset accession
        :type accession: str
        :param children: list of children accessions to add to the dataset
        :type children: list[str]
        """
        return self.invoke('addFiles', accession, children)

    def add_file_to_datasets(self, file_accession, dataset_accessions):
        """
        Add given file to several datasets.

        :param file_accession: file accession
        :type file_accession: str
        :param dataset_accessions: accessions of the datasets
        :type dataset_accessions: list[str]
        """
        return self.invoke('addFileToDatasets', file_accession, dataset_accessions)

    def remove_dataset_children(self, accession, children):
        """
        Remove children from dataset.

        :param accession: dataset accession
        :type accession: str
        :param children: list of children accessions to remove from the dataset
        :type children: list[str]
        """
        return self.invoke('removeFiles', accession, children)

    def merge_datasets(self, datasets, parent=None):
        """
        Create a new dataset from the given datasets.

        :param datasets: list of source datasets accessions
        :type datasets: list[str]
        :param parent: folder for the new dataset, 'My datasets' if not specified
        :type parent: str

        :return: accession of the created dataset
        :rtype: str
        """
        if parent is None:
            parent = self.__get_mydatasets_folder()

        return self.invoke('mergeDatasets', parent, datasets)

    def __get_mydatasets_folder(self):
        """
        Get default folder for datasets.

        :return: default dataset folder accession
        :rtype: str
        """
        return FilesUtil(self.connection).get_special_folder(SpecialFolders.MY_DATASETS)



# coding=utf-8

from __future__ import print_function

import json
import os
import sys

import jsonschema


SCHEMAS_DIR = 'schemas'

VALIDATION_ERR_MSG = '''Template validation failed

 Offending instance:
{instance}

 Error message:
{message}
'''


def load_json(file_path, schema_name):
    """
    Loads the JSON specified by the given path and validates it against a JSON schema.

    JSON schema must be located in the `schemas` directory of the current module.

    :param file_path: path to the JSON file
    :type file_path: str
    :param schema_name: name of the JSON schema file
    :type schema_name: str
    :return: decoded JSON instance
    :rtype: dict
    """
    base_dir = os.path.dirname(__file__)
    schema_path = os.path.join(base_dir, SCHEMAS_DIR, schema_name)
    with open(file_path) as data_file, open(schema_path) as schema_file:
        data = json.load(data_file)
        schema = json.load(schema_file)
        try:
            jsonschema.validate(data, schema)
        except jsonschema.exceptions.ValidationError as exc:
            print(VALIDATION_ERR_MSG.format(**exc.__dict__), file=sys.stderr)
            sys.exit(1)
        return data



from odm_sdk import Application


class SampleLinker(Application):
    """
    Application for linking data files to samples.

    It operates with the following concepts:

    1. A study is a dataset (collection) of samples.
    2. A sample is a file that contains common metainfo that can be attached to files with data.
    3. When linking data files and samples, data files must be uploaded and put into an upload
       dataset. This dataset simplifies operations on these files in Genestack and provides data
       versioning. Upload dataset is linked to the study.
    4. When uploading files to the upload dataset, they are put inside this dataset and
       initialized. Each file's metainfo will contain a link to the according sample.

    A typical workflow might look like this:

    1. A study with samples is created via the Study Design application inside Genestack.
    2. Study number is generated and exported via the Study Design API.
    3. An upload dataset is created and linked to the provided study.
    4. Files with data are uploaded, linked to samples and initialized via the 'import_data' method.
    5. If some data files are considered corrupted or invalid, they can be removed using the
       'unlink_data' method.
    6. When all required data files are uploaded, data can be made visible to others by releasing
       the upload dataset using the 'release' method.

    NOTE: This API is currently in Beta stage and is a subject to change, so no backwards
    compatibility is guaranteed at this point.
    """

    APPLICATION_ID = 'genestack/sample-linker'

    def create_upload_dataset(self, study_number, file_type, **kwargs):
        """
        Create a dataset that will later be used to hold uploaded data files.

        This method accepts additional parameters required for creating files inside Genestack.
        These parameters depend on the file type:

        * "ExpressionLevels": no additional parameters.

        Example:

        .. code-block:: python

            sample_linker.create_upload_dataset(
                study_number=1,
                file_type='ExpressionLevels'
            )

        Supported file types:

        * "ExpressionLevels": expression data
        * "MappedReadCounts": deprecated, use "ExpressionLevels" instead

        :param study_number: number of the study that contains samples for uploaded files.
        :type study_number: int
        :param file_type: type of files that will be uploaded
        :type file_type: str
        :param kwargs: additional options that are needed when creating a file. Options content
                       depends on the type of the created file.
        :return: accession of the created dataset
        :rtype: str
        """

        options = {
            key: value if isinstance(value, (list, tuple)) else [value]
            for key, value in kwargs.items()
        }
        return self.invoke('createUploadDataset', study_number, file_type, options)

    def import_data(self, samples, upload_dataset_accession):
        """
        Create data files inside the upload dataset and link them to the specified samples.

        Created files are initialized upon creation.

        NOTE: This method can only handle 100 files at a time, so in case
        of uploading more files than that they must be uploaded in batches of this size.

        Example:

        .. code-block:: python

            sample_linker.import_data(
                samples={
                    'sampleId1': ['http://data_url1', 'http://data_url2'],
                    'sampleId2': ['http://more.data']
                },
                upload_dataset_accession='GSF000123'
            )

        This call will return the following dictionary:

        .. code-block:: python

            {
                'sampleId1': ['GSF0001', 'GSF0002'],
                'sampleId2': ['GSF0003']
            }

        :param samples: mapping from sample id to a list of URLs that point to data.
        :type samples: dict[str, list[str]]
        :param upload_dataset_accession: accession of the upload dataset that will hold the created
                                         data files.
        :type upload_dataset_accession: str
        :return: mapping from sample id to a list of accessions of the created data files.
        :rtype: dict[str, list[str]]
        """
        return self.invoke('importData', samples, upload_dataset_accession)

    def unlink_data(self, file_accessions, upload_dataset_accession):
        """
        Remove uploaded data files from the given dataset and unlink them from their samples.
        Links to samples are always removed but actual files may not be removed from the system.

        Removing a file that isn't present in the dataset is a no-op and will not throw an
        exception.

        :param file_accessions: accessions of data files that should be unlinked.
        :type file_accessions: list[str]
        :param upload_dataset_accession: accession of the upload dataset that holds the provided
                                         data files.
        :type upload_dataset_accession: str
        """
        return self.invoke('unlinkData', file_accessions, upload_dataset_accession)

    def release(self, group_name, upload_dataset_accession):
        """
        Release the provided dataset. Releasing a dataset means that all data files are ready
        and can be shared with the outer world.

        This method is idempotent and can be run multiple times in case of errors.

        :param group_name: name of the group that the provided dataset will be shared with.
        :type group_name: str
        :param upload_dataset_accession: accession of the dataset that holds the uploaded data
                                         files.
        :type upload_dataset_accession: str
        """
        return self.invoke('release', group_name, upload_dataset_accession)



# coding=utf-8
import sys


class Color(object):
    """
    Class that contains constants for coloring the console output.
    """
    BLUE = 34
    GREEN = 32
    RED = 31


def colored(message, color):
    """
    Format the provided message with the given color if the standard output points to the console.
    Otherwise the message is returned as-is.

    :param message: text to format
    :type message: str
    :param color: color of the text (see :class:`format.Color`)
    :type color: int
    :return: str
    """
    if sys.platform == 'win32' or not sys.stdout.isatty():
        return message
    return '\033[%dm%s\033[0m' % (color, message)


def print_result(accession, name, type_name, started=False):
    """
    Print information about a created file using the following format:

    ``Created <type_name> <accession> / <name>``

    :param accession: accession of the created file
    :type accession: str
    :param name: name of the created file
    :type name: str
    :param type_name: type of the created file
    :type type_name: str
    :param started: if *True* the printed message will start with `Created and started`.
           This parameter is useful to indicate that the created file has also started initializing.
    """
    print('%s %s %s / %s' % (
        'Created and started' if started else 'Created',
        type_name,
        colored(accession, Color.GREEN),
        colored(name, Color.BLUE)
    ))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function

from odm_sdk import get_connection, make_connection_parser, FilesUtil, SpecialFolders


def get_template_by_name(args):
    connection = get_connection(args)
    files_util = FilesUtil(connection)

    def find(name, parent):
        return files_util.find_application_page_file_by_name(name, parent)

    folders_to_search = [files_util.get_public_folder(),
                         files_util.get_special_folder(SpecialFolders.CREATED)
                         ]
    for folder in folders_to_search:
        accession = find(args.template_name, folder)
        if accession:
            return accession
    return None


def main():
    parser = make_connection_parser()
    parser.add_argument('--template-name', dest='template_name',
                        default='Default Template',
                        help='template_name')
    args = parser.parse_args()
    accession = get_template_by_name(args)
    print('Template with name "%s" has accession: %s' % (args.template_name, accession))


if __name__ == '__main__':
    main()



from odm_sdk import Application, FilesUtil, GenestackException, Metainfo


class DataFlowEditor(Application):
    APPLICATION_ID = 'genestack/datafloweditor'

    def __init__(self, connection, application_id=None):
        Application.__init__(self, connection, application_id)
        self.__cache = {}

    def create_dataflow(self, accession, name=None):
        """
        Creates a data flow based on the file provenance of the specified file.
        The nodes of the data flow can be accessed by the accession of the corresponding files in the file provenance.

        :param accession: file accession
        :type accession: str
        :param name: data flow name
        :type name: str
        :return: accession of the created data flow file
        :rtype: str
        :raise GenestackException:
        """
        response = self.invoke('initializeApplicationState', 'createFromSources', accession)

        if response['type'] == 'newPage':
            accession = response['fileInfo']['accession']
        elif response['type'] == 'existingPages':
            # If file already exists we expect to get the last created file.
            # Existing page contains files from first to last (or MAX QUERY)
            # TODO: in case there are more files then MAX QUERY (100 ATM),
            # the last file in response will not be really last
            # (it is almost impossible use case, though)
            file_info = response['fileInfos'][-1]
            accession = file_info['accession']
        else:
            raise GenestackException("Unknown response type: %s" % response['type'])
        if name:
            FilesUtil(self.connection).replace_metainfo_string_value([accession], Metainfo.NAME, name)
        return accession

    def add_files(self, page_accession, node_accession, files):
        """
        Add files to a data flow node.

        :param page_accession: accession of data flow file
        :type page_accession: str
        :param node_accession: accession of origin file in node
        :type node_accession: str
        :param files: list of accessions of files to add to the node
        :type files: list
        :rtype: None
        """
        node = self.__get_node_by_accession(page_accession, node_accession)
        self.invoke('addFiles', files, node, page_accession)

    def clear_node(self, page_accession, node_accession):
        """
        Remove all files from a data flow node.

        :param page_accession: accession of data flow file
        :type page_accession: str
        :param node_accession: accession of origin file in node
        :type node_accession: str
        :rtype: None
        """
        node = self.__get_node_by_accession(page_accession, node_accession)
        self.invoke('clearFile', node, page_accession)

    def __get_graph(self, page_accession):
        """
        Cache graph, to avoid extra requests.
        """
        if page_accession not in self.__cache:
            self.__cache[page_accession] = self.invoke('getFlowData', page_accession)
        return self.__cache[page_accession]

    def __get_node_by_accession(self, page_accession, accession):
        """
        Return node id by its accession.
        """
        for node, node_data in self.__get_graph(page_accession)['fullGraph'].items():
            if accession in node_data['userData']['originalAccessions']:
                return node



from odm_sdk import get_connection, Metainfo
from odm_sdk.scripts.dictionaries.load_init_share_dictionaries import (load_dictionary, sharing,
                                                                                   initialization)

def upload_disease():

    # Upload Disease Ontology
    disease_ontology = {
        'name': 'Disease Ontology',
        'url': 'https://odm-init.s3.amazonaws.com/dictionaries/with_specific_metainfo/doid.owl',
        'description': 'An ontology for describing the classification of human diseases organized by etiology. '
                       'The original file is available at http://purl.obolibrary.org/obo/doid.owl'
    }
    connection = get_connection()
    disease_ontology_accession = load_dictionary(connection, disease_ontology)
    sharing(connection, [disease_ontology_accession])
    initialization(connection, [disease_ontology_accession])

    # Upload Disease Ontology Extended with reference to Disease Ontology
    metainfo = Metainfo()
    metainfo.add_file_reference("genestack.dictionary:parent", disease_ontology_accession)
    disease_ontology_extended = {
        'name': 'Disease Ontology Extended',
        'url': 'https://odm-init.s3.amazonaws.com/dictionaries/with_specific_metainfo/disease_extension.csv',
        'description': 'An ontology for describing the classification of human diseases organized by etiology.'
                       ' The original file is available at http://purl.obolibrary.org/obo/doid.owl'
    }
    disease_ontology_extended_accession = load_dictionary(connection, disease_ontology_extended, metainfo=metainfo)
    sharing(connection, [disease_ontology_extended_accession])
    initialization(connection, [disease_ontology_extended_accession])


def upload_gene_ontology():

    gene_ontology = {
        "name": "Gene Ontology",
        "url": "https://odm-init.s3.amazonaws.com/dictionaries/with_specific_metainfo/go.owl",
        "description": "A controlled vocabulary describing the gene functions according to three aspects: "
                       "biological process, molecular function and cellular component. "
                       "The original file is available at http://purl.obolibrary.org/obo/go.owl"
    }

    connection = get_connection()
    metainfo = Metainfo()
    metainfo.add_string(Metainfo.DATA_TYPE, 'Gene Ontology')
    gene_ontology_accession = load_dictionary(connection, gene_ontology, metainfo=metainfo)
    sharing(connection, [gene_ontology_accession])
    initialization(connection, [gene_ontology_accession])


def upload_gene_dictionaries():
    organisms_to_load = [
        'Bos taurus',
        'Danio rerio',
        'Homo sapiens',
        'Macaca fascicularis',
        'Mus musculus',
        'Oryctolagus cuniculus',
        'Rattus norvegicus',
        'Sus scrofa'
    ]
    connection = get_connection()
    dictionary_accessions = []

    for organism in organisms_to_load:
        gene_dict = {
            "name": organism + " genes",
            "url": "https://odm-init.s3.ap-south-1.amazonaws.com/dictionaries/gene-dictionaries/" +
                   organism.replace(" ", "+") + ".csv",
            "description": "Dictionary contains the following information about gene: Ensembl gene name and id, "
                           "Ensembl transcript name and id, Ensembl protein id, Reactome pathway name and id, "
                           "Entrez gene id, Uniprot symbol, HGNC gene id and description, GO term label and id."
        }

        metainfo = Metainfo()
        metainfo.add_string('genestack.dictionary:termType', 'gene')
        metainfo.add_string('Release', '99')
        metainfo.add_string('genestack.bio:organism', organism)
        gene_ontology_accession = load_dictionary(connection, gene_dict, metainfo=metainfo)
        dictionary_accessions.append(gene_ontology_accession)

    sharing(connection, dictionary_accessions)


def main():
    """
    Currently we don't support metainfo parsing via auxiliary scripts from json in proper way.
    In the future we have to add this possibility and rework script to using only CLI call.
    """
    upload_disease()
    upload_gene_ontology()
    upload_gene_dictionaries()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function

import sys

from odm_sdk import (GenestackServerException, get_connection,
                              make_connection_parser)

from odm_sdk.scripts.metainfo_templates.importers import TemplateImporter
from odm_sdk.scripts.metainfo_templates.importers import format as tf


def main():
    parser = make_connection_parser()
    parser.add_argument('paths', metavar='<path>', nargs='+',
                        help='one or more paths to configuration files')
    args = parser.parse_args()

    config_paths = args.paths
    print('Number of templates to be updated: %d' % len(config_paths))

    connection = get_connection(args)
    importer = TemplateImporter(connection)
    for path in config_paths:
        try:
            importer.create_from_config(path)
        except GenestackServerException as e:
            print(tf.colored("  Got an error from server:\n{}".format(e.message),
                             tf.Color.RED), file=sys.stderr)
            sys.exit(1)


if __name__ == '__main__':
    main()



import json
import os
import re
import sys
import time
from datetime import datetime
from io import BytesIO
from threading import Condition, Lock, Thread

from OpenSSL.SSL import SysCallError
from requests.exceptions import RequestException

from odm_sdk import GenestackException
from odm_sdk.utils import isatty

RETRY_ATTEMPTS = 5
RETRY_INTERVAL = 2  # seconds
NUM_THREADS = 5
CHUNK_SIZE = 1024 * 1024 * 5  # 5mb


class Chunk(object):
    def __init__(self, number, start, size, chunk_size, total_size, token, filename, path, chunk_count, launch_time):
        self.data = {
            'resumableChunkSize': chunk_size,
            'resumableType': '',
            'resumableTotalSize': total_size,
            'resumableIdentifier': token,
            'resumableFilename': filename,
            'resumableRelativePath': path,
            'resumableTotalChunks': chunk_count,
            'launchTime': launch_time,
            'resumableChunkNumber': number,
            'resumableCurrentChunkSize': size
            }

        self.start = start
        self.size = size

    def __str__(self):
        return "Chunk %s %s bytes for %s" % (self.data['resumableChunkNumber'], self.size,
                                             self.data['resumableRelativePath'])

    def get_file(self):
        container = BytesIO()
        with open(self.data['resumableRelativePath'], 'rb') as f:
            f.seek(self.start)
            container.write(f.read(self.size))
        container.seek(0)
        return container


class PermanentError(GenestackException):
    """
    If this exception is thrown, the upload will not be able to be resumed.
    """
    pass


def with_lock(method):
    """
    Execute method with lock. Instance should have lock object in lock attribute.
    """
    def wrapper(self, *args):
        with self.lock:
            return method(self, *args)
    return wrapper


class ChunkedUpload(object):
    def __init__(self, application, path, chunk_size=None):
        if chunk_size is None:
            chunk_size = CHUNK_SIZE
        if chunk_size <= 0:
            raise GenestackException("Chunk size should be positive")

        self.chunk_upload_url = '/application/uploadChunked/%s/unusedToken' % application.application_id
        self.connection = application.connection

        self.lock = Lock()
        self.__iterator_lock = Lock()
        self.__output_lock = Lock()

        self.__application_result = None
        self.__has_application_result = False
        self.__finished = False
        self.__error = None
        self.thread_counter = 0

        self.condition = Condition()

        modified = datetime.fromtimestamp(os.path.getmtime(path))
        total_size = os.path.getsize(path)

        # TODO change according to javascript token
        token = '{total_size}-{name}-{date}'.format(total_size=total_size,
                                                    name=re.sub('[^A-z0-9_-]', '_', os.path.basename(path)),
                                                    date=modified.strftime('%a_%b_%d_%Y_%H_%M_%S'))
        self.token = token
        self.path = path
        # Last chunk can be larger than CHUNK_SIZE but less then two chunks.
        # Example: CHUNK_SIZE = 2
        # file size 2 > 1 chunk
        # file size 3 > 1 chunk
        # file size 4 > 2 chunk
        # file size 5 > 2 chunk

        if total_size < chunk_size * 2:
            chunk_count = 1
        else:
            chunk_count = total_size / chunk_size

        self.total_size = total_size
        self.filename = os.path.basename(path)
        self.path = path
        self.chunk_count = chunk_count
        launch_time = int(time.time() * 1000)

        # import from here to avoid circular imports
        # TODO move progress functions to other module.
        if isatty():
            from .connection import TTYProgress
            self.progress = TTYProgress()
        else:
            from .connection import DottedProgress
            self.progress = DottedProgress(40)

        def _iterator():
            start = 0
            info = [chunk_size, total_size, token, self.filename, path, chunk_count, launch_time]
            for x in range(1, chunk_count + 1):
                if x == chunk_count:
                    current_chunk_size = self.total_size - start
                else:
                    current_chunk_size = chunk_size
                yield Chunk(x, start, current_chunk_size, *info)
                start += current_chunk_size

        self.iterator = _iterator()

    @property
    @with_lock
    def application_result(self):
        return self.__application_result

    @application_result.setter
    @with_lock
    def application_result(self, value):
        self.__application_result = value

    @property
    @with_lock
    def has_application_result(self):
        return self.__has_application_result

    @has_application_result.setter
    @with_lock
    def has_application_result(self, value):
        self.__has_application_result = value

    @property
    @with_lock
    def finished(self):
        return self.__finished

    @finished.setter
    @with_lock
    def finished(self, value):
        self.__finished = value

    @property
    @with_lock
    def error(self):
        return self.__error

    @error.setter
    def error(self, value):
        self.__error = value

    def __update_progress(self, update_size):
        with self.__output_lock:
            self.progress(self.filename, update_size, self.total_size)

    def __process_chunk(self, chunk):
        """
        Try to upload a chunk of data in several attempts.
        :param chunk:
        :return:
        """
        file_cache = None
        upload_checked = False
        error = None

        for attempt in range(RETRY_ATTEMPTS):
            # Check if chunk is already uploaded
            if not upload_checked:
                try:
                    response = self.connection.get_request(self.chunk_upload_url, params=chunk.data, follow=False)
                except RequestException as e:
                    error = str(e)
                    time.sleep(RETRY_INTERVAL)
                    continue

                if response.status_code == 200:
                    self.__update_progress(chunk.size)
                    return
                else:
                    upload_checked = True

            # try to upload chunk
            if file_cache is None:
                file_cache = chunk.get_file()

            file_cache.seek(0)
            try:
                response = self.connection.post_multipart(self.chunk_upload_url,
                                                          data=chunk.data,
                                                          files={'file': file_cache},
                                                          follow=False)
            except (RequestException, SysCallError) as e:
                # check that any type of connection error occurred and retry.
                time.sleep(RETRY_INTERVAL)
                error = str(e)
                if self.connection.debug:
                    sys.stderr.write('%s/%s attempt to upload %s failed. Connection error: %s\n' %
                                     (attempt + 1, RETRY_ATTEMPTS, chunk, error))
                continue
            # done without errors
            if response.status_code == 200:
                self.__update_progress(chunk.size)
                data = json.loads(response.text)

                if data.get('lastChunkUploaded', False):
                    self.application_result = data['result']
                    self.has_application_result = True
                    self.finished = True
                return

            error = "Got response with status code: %s" % response.status_code
            # permanent errors
            if 400 <= response.status_code < 600:
                self.finished = True
                try:
                    data = json.loads(response.text)
                    if isinstance(data, dict) and 'error' in data:
                        error = data['error']
                except ValueError:
                    pass
                self.error = error
                return
            # other network errors, try again
            time.sleep(RETRY_INTERVAL)
            continue

        self.error = error
        self.finished = True

    def upload(self):
        def do_stuff():
            """
            The daemon will look for uploads.
            The daemon will quit if one of the following conditions is met:
                - all chunks have been processed
                - someone set self.finished to True
                - the server said that the file upload was complete
                - a permanent error was raised (4xx, 5xx)
                - the number of RETRY_ATTEMPTS was exceeded for a single chunk
            """
            with self.condition:
                self.thread_counter += 1
            try:
                while not self.finished:  # daemon working cycle
                    try:
                        with self.__iterator_lock:
                            chunk = next(self.iterator)
                    except StopIteration:
                        return
                    self.__process_chunk(chunk)
            except Exception as e:
                self.error = str(e)
            finally:
                with self.condition:
                    self.thread_counter -= 1
                    self.condition.notify()

        threads = [Thread(target=do_stuff) for _ in range(min(NUM_THREADS, self.chunk_count))]
        [thread.setDaemon(True) for thread in threads]
        [thread.start() for thread in threads]

        with self.condition:
            while True:
                try:
                    self.condition.wait()
                except (KeyboardInterrupt, SystemExit):
                    self.error = 'Interrupted by user'
                    self.finished = True
                    break
                if not self.thread_counter:
                    break
        if self.has_application_result:
            return self.application_result
        else:
            error_message = self.error or 'file has been uploaded from another session'
            raise GenestackException('Fail to upload %s: %s' % (self.path, error_message))


def upload_by_chunks(application, path, chunk_size=None):
    return ChunkedUpload(application, path, chunk_size=chunk_size).upload()



#!/usr/bin/env python3

#
# Copyright (c) 2011-2022 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
#

import sys
import time
import requests
import argparse

EXPECTED_RESPONSE_CODE = 200
ALLOWED_WAIT_CYCLES = 120
TIMEOUT_IN_SECONDS = 10


def is_api_ready(odm_url):
    try:
        response = requests.get(
            url=f'{odm_url}/frontend/health',
        )
        return response.status_code
    except Exception:
        return 0


def main():
    arg_parser = argparse.ArgumentParser()

    arg_parser.add_argument(
        '--odm-url',
        type=str, help='host address', required=True
    )

    args = arg_parser.parse_args()

    wait_cycle = 0

    while wait_cycle < ALLOWED_WAIT_CYCLES:
        time.sleep(TIMEOUT_IN_SECONDS)
        if is_api_ready(args.odm_url):
            print(f'Host {args.odm_url} is ready!')
            sys.exit(0)
        else:
            wait_cycle += 1
            print(
                f'Iteration {wait_cycle}: {args.odm_url}/ '
                f'is not ready yet. Waiting another {TIMEOUT_IN_SECONDS} '
                'seconds to re-check.'
            )

    print('Waiting timeout exceeded. Host is not ready.')
    sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import argparse
import ftplib
import csv
import gzip
import os
import re
import shutil
import time

from collections import defaultdict
from subprocess import check_call

# This script combines code from initialise_raw_gene_dictionary.py and gene_dictionary_creation.py,
# previously located in unified repo. Look up their history there for reference.
# It primarily adapts initialise_raw_gene_dictionary.py, excluding the use of raw file metadata.
# Instead, it incorporates get_ensembl_organisms from gene_dictionary_creation.py.

MAIN_FILE_NAME = 'gene_ensembl__translation__main'
FIRST_COLUMN_NAMES = ['label', 'gene_symbol', 'ensembl_transcript_id', 'ensembl_protein_id', 'description']

# to comply with the Genestack CSV initializer which uses the default 'excel' dialect
ROW_ENTRY_SEPARATOR = ','
TERM_SEPARATOR = ','
BASE_ENSEMBL_FTP = 'ftp://ftp.ensembl.org/pub/release-99/mysql/ensembl_mart_99/'
ENSEMBL_SCHEMA_FILE = 'ensembl_mart_99.sql.gz'
BASE_NCBI_FTP = 'https://ftp.ncbi.nlm.nih.gov/'
NCBI_GENE_DATA_FOLDER = 'gene/DATA/'
NCBI_GENE_INFO_FILENAME = 'gene_info'
NCBI_GENE_INFO_FILENAME_GZ = 'gene_info.gz'


ORGANISMS_TO_LOAD_GENE_SYNONYMS = [
    'Bos taurus',
    'Canis familiaris',
    'Cricetulus griseus chok1gshd',
    'Danio rerio',
    'Homo sapiens',
    'Macaca fascicularis',
    'Mus musculus',
    'Oryctolagus cuniculus',
    'Rattus norvegicus',
    'Sus scrofa'
]

# Common files
DATASET_NAMES_TABLE_GZ = 'dataset_names.txt.gz'
COMMON_FILES = ['dataset_names']

# Organism specific files
DATABASE_FILES = [
    {
        'name': 'gene_ensembl__ox_reactome_gene__dm',
        'columns': {
            'description_1074': 'reactome_pathway_name',
            'display_label_1074': 'reactome_pathway_id'
        },
        # name of the column on which to perform the JOIN
        # (identical in this file and the create_files file)
        'join_key': 'gene_id_1020_key'
    },

    {
        'name': 'gene_ensembl__ox_entrezgene__dm',
        'columns': {
            'dbprimary_acc_1074': 'entrez_gene_id'
        },
        'join_key': 'gene_id_1020_key'
    },

    {
        'name': 'gene_ensembl__ox_uniprot_gn__dm',
        'columns': {
            'dbprimary_acc_1074': 'uniprot_symbol'
        },
        'join_key': 'gene_id_1020_key'
    },

    {
        'name': 'gene_ensembl__ox_hgnc__dm',
        'columns': {
            'description_1074': 'hgnc_description',
            'dbprimary_acc_1074': 'hgnc_id'
        },
        'join_key': 'gene_id_1020_key'
    },

    {
        'name': 'gene_ensembl__gene__main',
        'columns': {
            'name_1059': 'location',
        },
        'join_key': 'gene_id_1020_key'
    },

    {
        'name': 'gene_ensembl__ontology_go__dm',
        'columns': {
            'name_1006': 'go_term_label',
            'dbprimary_acc_1074': 'go_term_id'  # TODO what about GO category?
        },
        'join_key': ['transcript_id_1064_key', 'translation_id_1068_key']
    },

    {  # this is the create_files Ensembl database file
        'name': MAIN_FILE_NAME,
        'columns': {
            'stable_id_1070': 'ensembl_protein_id',
            'stable_id_1066': 'ensembl_transcript_id',
            'stable_id_1023': 'label',  # Ensembl gene ID
            'display_label_1074_r1': 'transcript_symbol',
            'display_label_1074': 'gene_symbol'
        }
    }
]


def get_ensembl_organisms():
    ensembl_ftp_host = 'ftp.ensembl.org'
    ensembl_ftp_mart_path = '/pub/release-99/mysql/ensembl_mart_99/'
    ensembl_ftp_parent_path = '/pub/release-99/mysql/'

    print("Retrieving list of organisms from Ensembl FTP...")
    ftp = ftplib.FTP(ensembl_ftp_host)
    ftp.login('anonymous', 'anonymous')

    # get the list of downloadable database dumps
    files_listing = ftp.nlst(ensembl_ftp_mart_path)
    to_upload = set(os.path.basename(path).split('_')[0] for path in files_listing
                    if os.path.basename(path).endswith('.txt.gz'))
    to_upload.discard('meta')
    to_upload.discard('dataset')

    # use the parent directory listing to match the names of these database dump files
    # to the full name of the corresponding organisms
    files_listing_parent_directory = ftp.nlst(ensembl_ftp_parent_path)
    organisms = {}
    for path in files_listing_parent_directory:
        split_name = os.path.basename(path).split('_')

        for i, part in enumerate(split_name):
            short_name = split_name[0][0] + split_name[i]
            if short_name in to_upload:
                display_name = ' '.join(map(str, split_name[0:(i + 1)]))
                display_name = display_name[0].upper() + display_name[1:]
                organisms[short_name] = display_name
                break

    print('\tTotal short names found: %s' % len(to_upload))
    for short_name in to_upload:
        if short_name not in organisms:
            print('\t[WARNING] No display name found for short name: %s' % short_name)
    print('\tTotal organisms found: %s' % len(organisms))
    return sorted(organisms.items())


def multiple_attempts(exception_to_raise, max_attempts=3, delay=3, backoff=1):
    """A decorator to retry a function call many times."""

    def _multiple_attempts(function):
        def wrapper(*args, **kwargs):
            n_attempts = 0
            next_attempt_delay = delay
            while True:
                try:
                    return function(*args, **kwargs)
                except Exception as e:
                    n_attempts += 1
                    if n_attempts > max_attempts:
                        if exception_to_raise:
                            print("Received exception: %s" % e)
                            raise exception_to_raise
                        raise
                    time.sleep(next_attempt_delay)
                    next_attempt_delay *= backoff
                    print("Retrying...")

        return wrapper
    return _multiple_attempts


def update_list_dict(dic1, dic2):
    # Update list-valued dictionary like this:
    # { 1: [2,3], 2: [5] } + {1: [4,6], 3: [2]} ==> {1: [2,3,4,6], 2: [5], 3: [2]}
    # (in contrast, the method dict.update will just overwrite instead of merging)
    for key, values in dic2.items():
        dic1[key] += values


def unique_list(lst):
    # remove duplicate entries from a list whilst preserving order:
    # [1, 5, 3, 5, 2, 1] -> [1, 5, 3, 2]
    already_seen = set()
    output = []
    for elem in lst:
        if elem not in already_seen:
            already_seen.add(elem)
            output.append(elem)
    return output


def quote_item_if_contains_comma(item):
    """If term contains comma itself, it shoud be quoted before writing into .csv."""
    return '"' + item + '"' if ',' in item else item


def prepare_cell(items):
    """
    We construct cell from list of items (str) that quoted and joined via separator.
    """
    return TERM_SEPARATOR.join(quote_item_if_contains_comma(item) for item in items)


def prepare_csv_row(row):
    """
    row = ['1', ['2', '3'], ['4', '5, 6', '7']] ==> ['1', '2,3', '4,"5, 6",7'].
    Such format of csv row is expected by genestack OntologyCsvInitializer.java class
    :param row: preliminary representation of the csv row
    :type row: list[list[str]]
    :return: list[str]
    """
    return [prepare_cell(items) for items in row]


@multiple_attempts(exception_to_raise=None, max_attempts=10, delay=1, backoff=2)
def download_file(url, local_target):
    # For some reason downloading with urllib/urllib2 is very slow:
    # urllib.urlretrieve(url, local_target)
    print('\tDownloading: %s' % local_target)

    tmp_name = '%s.download' % local_target
    check_call("curl -s '%s' --output '%s'" % (url, tmp_name), shell=True)

    if os.path.isfile(local_target):
        os.remove(local_target)
    shutil.move(tmp_name, local_target)


def parse_ensembl_mart_schema():
    schema_dict = {}
    print('Parsing Ensembl mart schema...')
    local_file = ENSEMBL_SCHEMA_FILE
    download_file(BASE_ENSEMBL_FTP + ENSEMBL_SCHEMA_FILE, local_file)
    selected_file_names = set(descriptor['name'] for descriptor in DATABASE_FILES)

    with gzip.open(local_file, mode='rt', encoding='utf-8') as sql_file:
        table_entries = {}
        skip_current_table = False
        col_counter = -1
        table_name = None

        for line in sql_file:
            # if we get to a new table declaration
            if line.startswith('CREATE TABLE'):
                col_counter = 0
                # if a previous table was parsed, store the parsed info in `schema_dict`
                if len(table_entries) > 0:
                    schema_dict[table_name] = table_entries
                    table_entries = {}

                table_name = re.search(r'CREATE TABLE `(.*)`', line).group(1)
                table_type = table_name.split('_', 1)[1]  # remove the organism
                skip_current_table = (table_type not in selected_file_names) and (table_name not in COMMON_FILES)

            elif skip_current_table:
                continue
            else:
                match = re.search(r'^\s+`(.*)`', line)
                if match is None:
                    continue
                col_name = match.group(1)
                # at the moment, we store the index mapping for *all* columns
                table_entries[col_name] = col_counter
                col_counter += 1

        if len(table_entries) > 0:
            schema_dict[table_name] = table_entries

    print('\tTotal tables used: %s' % len(schema_dict))

    return schema_dict


# TODO NCBI gene synonyms should be added to dictionary before initial write on disk
def add_gene_synonyms_from_ncbi(result_filename, taxonomy_id):
    """
    Add gene synonyms extracted from NCBI `gene_info` file to result .csv file
    :param result_filename: result .csv file where gene synonyms should be added
    :param taxonomy_id: taxonomy id of the organism used to extract organism specific rows from NCBI `gene_info` file
    :return: None
    """
    gene_synonyms_dict = load_gene_synonyms_dict(taxonomy_id)
    write_result_with_synonyms(result_filename, gene_synonyms_dict)


def load_gene_synonyms_dict(taxonomy_id):
    gene_synonyms_dict = {}
    with gzip.open(NCBI_GENE_INFO_FILENAME_GZ, mode='rt', encoding='utf-8') as gene_info_file:
        reader = csv.DictReader(gene_info_file, delimiter='\t')
        for row in reader:
            if row['#tax_id'] == taxonomy_id:
                gene_synonyms = row['Synonyms'].split('|') if row['Synonyms'] != '-' else []
                if row['Symbol_from_nomenclature_authority'] != '-':
                    gene_symbol = row['Symbol_from_nomenclature_authority']
                    if gene_symbol != row['Symbol']:
                        gene_synonyms.append(row['Symbol'])
                else:
                    gene_symbol = row['Symbol']
                gene_synonyms_dict[gene_symbol] = TERM_SEPARATOR.join(gene_synonyms)
    return gene_synonyms_dict


def load_taxonomy_dict(ensembl_column_mappings):
    species_col_index = ensembl_column_mappings['dataset_names']['name']
    tax_id_col_index = ensembl_column_mappings['dataset_names']['tax_id']

    download_file(BASE_ENSEMBL_FTP + DATASET_NAMES_TABLE_GZ, DATASET_NAMES_TABLE_GZ)
    with gzip.open(DATASET_NAMES_TABLE_GZ, mode='rt', encoding='utf-8') as dataset_names_file:
        reader = csv.reader(dataset_names_file, delimiter='\t')
        return {row[species_col_index]: row[tax_id_col_index] for row in reader}


def read_result_file(result_filename):
    with open(result_filename, 'rt', encoding='utf-8') as result_file:
        reader = csv.reader(result_file, delimiter=ROW_ENTRY_SEPARATOR)
        header = next(reader)
        result_file_content = list(reader)
    return header, result_file_content


def write_result_with_synonyms(result_filename, gene_synonyms_dict):
    header, result_file_content = read_result_file(result_filename)
    gene_symbol_index = header.index('gene_symbol')
    header.append('gene_synonyms')

    with open(result_filename, 'w') as result_file:
        writer = csv.writer(result_file, delimiter=ROW_ENTRY_SEPARATOR)
        writer.writerow(header)
        for row in result_file_content:
            gene_symbol = row[gene_symbol_index]
            row.append(gene_synonyms_dict.get(gene_symbol, ''))
            writer.writerow(row)


def get_foreign_key(col_names_to_indices, join_key):
    if not isinstance(join_key, list):
        join_key = [join_key]
    for possible_key in join_key:
        if possible_key in col_names_to_indices:
            return col_names_to_indices[possible_key]


def main():
    # Set up command-line argument parsing
    parser = argparse.ArgumentParser(description='Generates source files for gene dictionaries.')
    parser.add_argument('-ncbi', action='store_true', help='Generate only dictionaries with NCBI synonyms.')

    args = parser.parse_args()
    ncbi_patched_only = args.ncbi

    # download ncbi gene_info file in advance
    download_file(BASE_NCBI_FTP + NCBI_GENE_DATA_FOLDER + NCBI_GENE_INFO_FILENAME_GZ, NCBI_GENE_INFO_FILENAME_GZ)
    ensembl_column_mappings = parse_ensembl_mart_schema()
    taxonomy_dict = load_taxonomy_dict(ensembl_column_mappings)
    output_columns = FIRST_COLUMN_NAMES
    for dic in DATABASE_FILES:
        for value in dic['columns'].values():
            if value not in FIRST_COLUMN_NAMES:
                output_columns.append(value)

    for species, organism_name in get_ensembl_organisms():
        if ncbi_patched_only and organism_name not in ORGANISMS_TO_LOAD_GENE_SYNONYMS:
            continue
        secondary_files_data = []

        # dictionary: Ensembl gene id -> dictionary of gene info
        final_data = defaultdict(lambda: defaultdict(list))

        for file_descriptor in DATABASE_FILES:
            # for each of the secondary files for the current species
            # (i.e. not the create_files Ensembl file),
            # store the contents of the file in a dict where keys are the foreign DB keys,
            # and values are lists of dictionaries corresponding to the matching rows in the file.
            # Then, when we get to the create_files file (MAIN_FILE),
            # we combine its data with the data we stored in
            # memory for all the other files and write to the output CSV file

            is_main_file = file_descriptor['name'] == MAIN_FILE_NAME
            file_name = '%s_%s.txt.gz' % (species, file_descriptor['name'])
            file_data = defaultdict(list)
            col_names_to_indices = ensembl_column_mappings.get('%s_%s' % (species, file_descriptor['name']))

            if col_names_to_indices is None:
                print('\tFile `%s_%s` not found in SQL schema' % (species, file_descriptor['name']))
                continue

            if not is_main_file:
                foreign_key_id = get_foreign_key(col_names_to_indices, file_descriptor['join_key'])

            local_name = file_name
            download_file(BASE_ENSEMBL_FTP + file_name, local_name)

            if is_main_file:
                print('\tWriting species dictionary...')

            with gzip.open(local_name, mode='rt', encoding='cp1251') as txt_file:
                csv_reader = csv.reader(txt_file, delimiter='\t')
                for entry in csv_reader:
                    row_dictionary = defaultdict(list)
                    for input_name, output_name in file_descriptor['columns'].items():
                        col_index = col_names_to_indices[input_name]
                        if col_index < len(entry) and entry[col_index] and entry[col_index] != '\\N':
                            row_dictionary[output_name].append(entry[col_index])

                    if is_main_file:
                        if 'label' not in row_dictionary:
                            # if the row doesn't correspond to a "ENSGxxxxxx", skip it
                            continue
                        # add to the current row dictionary the data collected from the other 2ary files
                        for j, secondary_file_data in enumerate(secondary_files_data):
                            # do the JOIN operation
                            foreign_key_value = \
                                entry[get_foreign_key(col_names_to_indices, DATABASE_FILES[j]['join_key'])]
                            if foreign_key_value in secondary_file_data:
                                for secondary_row_dict in secondary_file_data[foreign_key_value]:
                                    # each `secondary_row_dict` corresponds to a row from the secondary file
                                    # which matches the current gene
                                    # (there can be many such rows, e.g. for GO terms)
                                    update_list_dict(row_dictionary, secondary_row_dict)

                        # update the primary file's dictionary
                        ensembl_gene_id = row_dictionary['label'][0]
                        update_list_dict(final_data[ensembl_gene_id], row_dictionary)
                    else:
                        # not the create_files file
                        # It is not main file so foreign_key_id is assigned
                        # noinspection PyUnboundLocalVariable
                        foreign_key_value = entry[foreign_key_id]
                        file_data[foreign_key_value].append(row_dictionary)

            if not is_main_file:
                secondary_files_data.append(file_data)

        # write to output dictionary
        result_name = '%s.csv' % organism_name

        with open(os.path.join(result_name), 'w') as species_file:
            output_writer = csv.writer(species_file, delimiter=ROW_ENTRY_SEPARATOR)
            output_writer.writerow(output_columns)
            for row_dict in final_data.values():
                row = prepare_csv_row(unique_list(row_dict[key]) for key in output_columns)
                output_writer.writerow(row)

        # Add gene synonyms here
        if organism_name in ORGANISMS_TO_LOAD_GENE_SYNONYMS:
            add_gene_synonyms_from_ncbi(os.path.join(result_name), taxonomy_dict[species])


if __name__ == "__main__":
    main()
